# -*- coding: utf-8 -*-
# Copyright 2018-2019 releng-tool

# Implementation in this file is considered "API safe". There is a strong
# attempt to prevent the changing of the following classes, methods, etc. to
# prevent compatibility issues as both this tool and extensions (if any) evolve.

class RelengRegistryInterface(object):
    """
    interface of the registry passed into an extension's setup stage

    This is the public interface of the releng-tool's registry which outlines
    methods which can be invoked by an extension during the setup stage. When an
    extension defines "def releng_setup(app):" the ``app`` value will be an
    instance of this interface (``RelengRegistryInterface``). During the setup
    call, an extension may call any of the defined methods using their ``app``
    context.
    """

    def add_extract_type(self, name, handler):
        """
        register extraction support with a given name for this extension

        An extension will invoke this method will attempting to support a custom
        extraction method for a release engineering process. The extraction
        type is identified by the provided ``name`` value. Packages being
        processed with VCS type paired with a extension-defined fetch type or a
        package extraction override will be processed through the provided
        ``handler``.

        An extension must define a proper ``name`` for the extract-type:

         - The name is a string value.
         - The name must be prefixed with ``ext-``.
         - The name should be all lowercase.
         - The dash character ``-`` is the recommended separator.

        The ``handler`` must be compatible with the implementation defined in
        ``RelengExtractExtensionInterface``. There is no requirement that the
        handler needs to inherit this interface.

        Additional notes:

         - The releng-tool process may instantiation one or more of the provided
            handler types.
         - The first extension loaded with a custom extract-type will take
            precedence over other extensions.

        Args:
            name: the name of this extract type
            handler: the extract-type class handler

        Raises:
            RelengInvalidSetupException: raised when the provided ``name`` or
                ``handler`` values are not supported by the releng-tool process
        """

    def add_fetch_type(self, name, handler):
        """
        register fetch support with a given name for this extension

        An extension will invoke this method will attempting to support a custom
        fetch method for a release engineering process. The fetch type is
        identified by the provided ``name`` value. Packages being processed with
        a matching VCS type will be processed through the provided ``handler``.

        An extension must define a proper ``name`` for the fetch-type:

         - The name is a string value.
         - The name must be prefixed with ``ext-``.
         - The name should be all lowercase.
         - The dash character ``-`` is the recommended separator.

        The ``handler`` must be compatible with the implementation defined in
        ``RelengFetchExtensionInterface``. There is no requirement that the
        handler needs to inherit this interface.

        Additional notes:

         - The releng-tool process may instantiation one or more of the provided
            handler types.
         - The first extension loaded with a custom fetch-type will take
            precedence over other extensions.

        Args:
            name: the name of this fetch type
            handler: the fetch-type class handler

        Raises:
            RelengInvalidSetupException: raised when the provided ``name`` or
                ``handler`` values are not supported by the releng-tool process
        """

    def add_package_type(self, name, handler):
        """
        register package support with a given name for this extension

        An extension will invoke this method will attempting to support a custom
        package methods for a release engineering process. The package type is
        identified by the provided ``name`` value. Packages being processed with
        a matching type will be processed through the provided ``handler`` for
        each package stage (configuration, building, etc.).

        An extension must define a proper ``name`` for the package-type:

         - The name is a string value.
         - The name must be prefixed with ``ext-``.
         - The name should be all lowercase.
         - The dash character ``-`` is the recommended separator.

        The ``handler`` must be compatible with the implementation defined in
        ``RelengPackageExtensionInterface``. There is no requirement that the
        handler needs to inherit this interface.

        Additional notes:

         - The releng-tool process may instantiation one or more of the provided
            handler types.
         - The first extension loaded with a custom package-type will take
            precedence over other extensions.

        Args:
            name: the name of this package type
            handler: the package-type class handler

        Raises:
            RelengInvalidSetupException: raised when the provided ``name`` or
                ``handler`` values are not supported by the releng-tool process
        """

    def require_version(self, version):
        """
        perform a required-version check

        Enables an extension to explicitly check for a required releng-tool
        version before being loaded. Invoking this method with a
        dotted-separated ``version`` string, the string will be parsed and
        compared with the running releng-tool version. If the required version
        is met, this method will have no effect. In the event that the required
        version is not met, the exception ``RelengVersionNotSupportedException``
        will be raised. The extension implementation does not need to explicitly
        handle this exception (as it is caught by the registry loading process).

        Args:
            version: dotted-separated version string

        Raises:
            RelengVersionNotSupportedException: raised when the required version
                for this extension is not met
        """

# ##############################################################################

class RelengPackageOptions(object):
    """
    releng package-type options

    Provides a series of options from the releng process into a package-type
    handler. A handler's ``configure``, ``build``, and ``install`` methods will
    be passed options to react on.

    Attributes:
        def_dir: the directory holder the package definition
        env: package's environment variables
        ext: extension (pass-through) options
        name: the name of the package being processed
        version: the version of the package
    """
    def __init__(self):
        self.def_dir = None
        self.env = None
        self.ext = {}
        self.name = None
        self.version = None

class RelengConfigureOptions(RelengPackageOptions):
    """
    releng configure-type options

    Provides a series of options from the releng process into a configure-type
    handler. A handler's ``configuration`` method will be passed options to
    react on.

    Attributes:
        build_dir: directory for a package's buildable content
        build_output_dir: build output directory for the package process
        conf_defs: command line definitions to apply
        conf_env: environment variables to use
        conf_opts: command line options to apply
        host_dir: directory container for host tools
        install_type: installation type (host, staging, target or images)
        jobs: number of calculated jobs to allow at a given time
        jobsconf: number of jobs to allow at a given time (0: auto)
        prefix: prefix for system root (if applicable)
        staging_dir: directory container for staged content
        symbols_dir: directory container for symbols content
        target_dir: directory container for target content
    """
    def __init__(self):
        super(RelengConfigureOptions, self).__init__()
        self.build_dir = None
        self.build_output_dir = None
        self.conf_defs = None
        self.conf_env = None
        self.conf_opts = None
        self.host_dir = None
        self.install_type = None
        self.jobs = 1
        self.jobsconf = 0
        self.prefix = None
        self.staging_dir = None
        self.symbols_dir = None
        self.target_dir = None

class RelengBuildOptions(RelengPackageOptions):
    """
    releng build-type options

    Provides a series of options from the releng process into a build-type
    handler. A handler's ``build`` method will be passed options to react on.

    Attributes:
        build_defs: command line definitions to apply
        build_dir: directory for a package's buildable content
        build_env: environment variables to use
        build_opts: command line options to apply
        build_output_dir: build output directory for the package process
        host_dir: directory container for host tools
        jobs: number of calculated jobs to allow at a given time
        jobsconf: number of jobs to allow at a given time (0: auto)
        prefix: prefix for system root (if applicable)
        staging_dir: directory container for staged content
        symbols_dir: directory container for symbols content
        target_dir: directory container for target content
    """
    def __init__(self):
        super(RelengBuildOptions, self).__init__()
        self.build_defs = None
        self.build_dir = None
        self.build_env = None
        self.build_opts = None
        self.build_output_dir = None
        self.host_dir = None
        self.jobs = 1
        self.jobsconf = 0
        self.prefix = None
        self.staging_dir = None
        self.symbols_dir = None
        self.target_dir = None

class RelengExtractOptions(RelengPackageOptions):
    """
    releng extract-type options

    Provides a series of options from the releng process into a extract-type
    handler. A handler's ``extract`` method will be passed options to react on.

    Attributes:
        cache_dir: directory to store cached content (if applicable)
        cache_file: file to store cached content (if applicable)
        revision: revision to use to fetch from source control
        strip_count: strip-count for package extraction (if applicable)
        work_dir: the working directory
    """
    def __init__(self):
        super(RelengExtractOptions, self).__init__()
        self.cache_dir = None
        self.cache_file = None
        self.revision = None
        self.strip_count = 1
        self.work_dir = None

class RelengFetchOptions(RelengPackageOptions):
    """
    releng fetch-type options

    Provides a series of options from the releng process into a fetch-type
    handler. A handler's ``fetch`` method will be passed options to react on.

    Attributes:
        cache_dir: directory to store cached content (if applicable)
        cache_file: file to store cached content (if applicable)
        fetch_opts: command line options to apply (if applicable)
        ignore_cache: whether or not there is a hint to ignore cache information
        revision: revision to use to fetch from source control
        site: the site (uri) to acquire a package's resources
    """
    def __init__(self):
        super(RelengFetchOptions, self).__init__()
        self.cache_dir = None
        self.cache_file = None
        self.fetch_opts = None
        self.ignore_cache = None
        self.revision = None
        self.site = None

class RelengInstallOptions(RelengPackageOptions):
    """
    releng install-type options

    Provides a series of options from the releng process into a install-type
    handler. A handler's ``installation`` method will be passed options to react
    on.

    Attributes:
        build_dir: directory for a package's buildable content
        build_output_dir: build output directory for the package process
        cache_file: location to cache file (if applicable)
        dest_dirs: list of directories to install to
        host_dir: directory container for host tools
        images_dir: directory container for (final) images
        install_defs: command line definitions to apply
        install_env: environment variables to use
        install_opts: command line options to apply
        install_type: installation type
        prefix: prefix for system root (if applicable)
        staging_dir: directory container for staged content
        symbols_dir: directory container for symbols content
        target_dir: directory container for target content
    """
    def __init__(self):
        super(RelengInstallOptions, self).__init__()
        self.build_dir = None
        self.build_output_dir = None
        self.cache_file = None
        self.dest_dirs = None
        self.host_dir = None
        self.images_dir = None
        self.install_defs = None
        self.install_env = None
        self.install_opts = None
        self.install_type = None
        self.prefix = None
        self.staging_dir = None
        self.symbols_dir = None
        self.target_dir = None

# ##############################################################################

class RelengExtensionInterface(object):
    """
    interface to implement a custom fetch-type

    Extensions wishing to define a custom fetch type can implement this
    interface and register it (see ``add_fetch_type``) during the extension's
    setup stage. This will allow an project support custom VCS-types defined in
    package definitions (for example "<PKG>_VCS_TYPE='ext-myawesomefetchtype'").
    """

class RelengExtractExtensionInterface(RelengExtensionInterface):
    """
    interface to implement a custom fetch-type

    Extensions wishing to define a custom fetch type can implement this
    interface and register it (see ``add_fetch_type``) during the extension's
    setup stage. This will allow an project support custom VCS-types defined in
    package definitions (for example "<PKG>_VCS_TYPE='ext-myawesomefetchtype'").
    """

    def extract(self, name, opts):
        """
        handle a custom extract operation

        When a package attempts to be "extracted" and VCS-type is defined to the
        custom extraction name, this method will be invoked. Package information
        to perform the extraction is provided by the ``opts`` argument. The main
        goal of this operation is to populate the provided work or "build"
        directory (same as the working directory) based off the package's cache
        directory or file (which varies base don the package type). On
        completion, this method should return ``True`` to indicate a successful
        extraction. If an error occurs, this method should return ``False``. A
        failed extraction operation is not required to clean the work directory.

        The ``extract`` method may be called multiple times throughout the
        lifetime of a releng-tool process (i.e. once per project being
        extracted).

        Args:
            name: the name of this extraction type
            opts: the extract options (see ``RelengExtractOptions``)

        Returns:
            ``True`` on successful extraction; otherwise ``False``
        """
        return False

class RelengFetchExtensionInterface(RelengExtensionInterface):
    """
    interface to implement a custom fetch-type

    Extensions wishing to define a custom fetch type can implement this
    interface and register it (see ``add_fetch_type``) during the extension's
    setup stage. This will allow an project support custom VCS-types defined in
    package definitions (for example "<PKG>_VCS_TYPE='ext-myawesomefetchtype'").
    """

    def fetch(self, name, opts):
        """
        handle a custom fetch operation

        When a package attempts to be "fetched" and VCS-type is defined to the
        custom fetch name, this method will be invoked. Package information to
        perform the fetch on is provided by the ``opts`` argument. The main goal
        of this operation is to fetch content from a provided site into either
        a cache file or a cache directory. On completion, this method should
        return either the cache file or cache directory depending on which has
        been populated from the fetch operation. The returned catch value should
        match the same value provided in the fetch options: ``opts.cache_file``
        if the fetch operation populates a cached file or ``opts.cache_dir`` if
        the fetch operation populates a cache directory. If the fetch operation
        returns a ``None`, this is an indication that the fetch operation has
        failed. A failed fetch operation is not required to clean the work
        directory; however, it is the responsibility of the custom fetch type to
        manage the state of the ``cache_dir`` (if applicable) and other external
        resources it may managed.

        The ``fetch`` method may be called multiple times throughout the
        lifetime of a releng-tool process (i.e. once per project being fetched).

        Args:
            name: the name of this fetch type
            opts: the fetch options (see ``RelengFetchOptions``)

        Returns:
            ``True`` on successful fetch; otherwise ``False``
        """
        return False

class RelengPackageExtensionInterface(RelengExtensionInterface):
    """
    interface to implement a custom fetch-type

    Extensions wishing to define a custom fetch type can implement this
    interface and register it (see ``add_fetch_type``) during the extension's
    setup stage. This will allow an project support custom VCS-types defined in
    package definitions (for example "<PKG>_VCS_TYPE='ext-myawesomefetchtype'").
    """

    def build(self, name, opts):
        """
        handle a custom build operation

        When a package reaches the building stage and package-type is defined
        to the custom package-type name, this method will be invoked. Package
        information to perform the build on is provided by the ``opts``
        argument. The main goal of this operation is to complete the build stage
        of a specific package in the provided work directory (same as the
        working directory). On completion, this method should return ``True`` to
        indicate a successful build. If an error occurs, this method should
        return ``False``. A failed build operation is not required to clean the
        work directory; however, the implementation should (attempt) to support
        a build request on a package which may have previously failed (i.e.
        build options or sources may have been changed to result in a successful
        build).

        The ``build`` method may be called multiple times throughout the
        lifetime of a releng-tool process (i.e. once per project being built).

        Args:
            name: the name of this package type
            opts: the build options (see ``RelengPackageOptions``)

        Returns:
            ``True`` on a successful build; otherwise ``False``
        """
        return False

    def configure(self, name, opts):
        """
        handle a custom configure operation

        When a package reaches the configuration stage and package-type is
        defined to the custom package-type name, this method will be invoked.
        Package information to perform the configuration on is provided by the
        ``opts`` argument. The main goal of this operation is to complete the
        configuration stage of a specific package in the provided work directory
        (same as the working directory). On completion, this method should
        return ``True`` to indicate a successful configuration. If an error
        occurs, this method should return ``False``. A failed configuration
        operation is not required to clean the work directory; however, the
        implementation should (attempt) to support a configuration request on a
        package which may have previously failed (i.e. configuration options or
        sources may have been changed to result in a successful configuration).

        The ``configure`` method may be called multiple times throughout the
        lifetime of a releng-tool process (i.e. once per project being
        configured).

        Args:
            name: the name of this package type
            opts: the configuration options (see ``RelengPackageOptions``)

        Returns:
            ``True`` on a successful configuration; otherwise ``False``
        """
        return False

    def install(self, name, opts):
        """
        handle a custom install operation

        When a package reaches the installation stage and package-type is
        defined to the custom package-type name, this method will be invoked.
        Package information to perform the installation on is provided by the
        ``opts`` argument. The main goal of this operation is to complete the
        install stage of a specific package in the provided work directory
        (same as the working directory). On completion, this method should
        return ``True`` to indicate a successful installation. If an error
        occurs, this method should return ``False``. A failed install operation
        is not required to clean the work directory; however, the implementation
        should (attempt) to support an install request on a package which may
        have previously failed (i.e. install options or sources may have been
        changed to result in a successful installation).

        The ``install`` method may be called multiple times throughout the
        lifetime of a releng-tool process (i.e. once per project being
        installed).

        Args:
            name: the name of this package type
            opts: the install options (see ``RelengPackageOptions``)

        Returns:
            `True`` on a successful installation; otherwise ``False``
        """
        return False

# ##############################################################################

class RelengInvalidSetupException(Exception):
    """
    exception raised when a loading extension has a setup issue
    """

class RelengVersionNotSupportedException(Exception):
    """
    exception raised when an extension does not support the releng-tool version
    """



# -*- coding: utf-8 -*-
# Copyright 2018 releng-tool

def releng_setup(app):
    seed_ext_type = RelengSeedExtension
    app.add_fetch_type('ext-seed', seed_ext_type)
    app.add_extract_type('ext-seed', seed_ext_type)
    app.add_package_type('ext-seed', seed_ext_type)

class RelengSeedExtension:
    def fetch(self, name, opts):
        print('(seed) package', opts.name, 'has been fetched')

        try:
            open(opts.cache_file, 'w').close()
            return opts.cache_file
        except IOError:
            return None

    def extract(self, name, opts):
        print('(seed) package', opts.name, 'has been extracted')
        return True

    def configure(self, name, opts):
        print('(seed) package', opts.name, 'has been configured')
        return True

    def build(self, name, opts):
        print('(seed) package', opts.name, 'has been built')
        return True

    def install(self, name, opts):
        print('(seed) package', opts.name, 'has been installed')
        return True



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

from releng_tool.packages.exceptions import RelengToolMissingPackageScript
from tests import prepare_testenv
from tests import run_testenv
import os
import unittest

class TestEngineRunActions(unittest.TestCase):
    def test_engine_run_actions_invalid_init(self):
        config = {
            'action': 'init',
        }

        # init should fail if we already have a project
        rv = run_testenv(config=config, template='minimal')
        self.assertFalse(rv)

    def test_engine_run_actions_valid_init(self):
        config = {
            'action': 'init',
        }

        with prepare_testenv(config=config) as engine:
            root_dir = engine.opts.root_dir
            releng_script = os.path.join(root_dir, 'releng')
            self.assertFalse(os.path.exists(releng_script))

            rv = engine.run()
            self.assertTrue(rv)
            self.assertTrue(os.path.exists(releng_script))

    def test_engine_run_actions_unknown_package(self):
        config = {
            'action': 'unknown-package',
        }

        with self.assertRaises(RelengToolMissingPackageScript):
            run_testenv(config=config, template='minimal')

    def test_engine_run_actions_valid_action(self):
        config = {
            'action': 'fetch',
        }

        rv = run_testenv(config=config, template='minimal')
        self.assertTrue(rv)

    def test_engine_run_actions_valid_package(self):
        config = {
            'action': 'minimal',
        }

        rv = run_testenv(config=config, template='minimal')
        self.assertTrue(rv)

    def test_engine_run_actions_valid_package_subaction(self):
        config = {
            'action': 'minimal-fetch',
        }

        rv = run_testenv(config=config, template='minimal')
        self.assertTrue(rv)



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from glob import glob
from releng_tool.tool.patch import PATCH
from releng_tool.util.log import err
from releng_tool.util.log import note
from releng_tool.util.log import verbose
from runpy import run_path
import os
import sys

#: filename of the script to execute custom patching operations (if any)
PATCH_SCRIPT = 'patch'

def stage(engine, pkg, script_env):
    """
    handles the patching stage for a package

    With a provided engine and package instance, the patching stage will be
    processed.

    Args:
        engine: the engine
        pkg: the package being patched
        script_env: script environment information

    Returns:
        ``True`` if the patching stage is completed; ``False`` otherwise
    """

    if pkg.is_internal:
        # internal packages in local sources mode do not have a patch stage
        if engine.opts.local_srcs:
            return True

        # internal packages in development mode that specify a development
        # revision will not perform the patch stage
        if engine.opts.devmode and pkg.has_devmode_option:
            return True

    note('patching {}...', pkg.name)
    sys.stdout.flush()

    if pkg.build_subdir:
        build_dir = pkg.build_subdir
    else:
        build_dir = pkg.build_dir

    patch_script_filename = '{}-{}'.format(pkg.name, PATCH_SCRIPT)
    patch_script = os.path.join(pkg.def_dir, patch_script_filename)
    if os.path.isfile(patch_script):
        try:
            run_path(patch_script, init_globals=script_env)

            verbose('patch script executed: ' + patch_script)
        except Exception as e:
            err('error running patch script: \n'
                '    {}', patch_script, e)
            return False

    # find all patches in the package's folder, sort and apply each
    patch_glob = os.path.join(pkg.def_dir, '*.patch')
    patches = glob(patch_glob)
    if patches:
        patches = sorted(patches)
        if not PATCH.exists():
            err('unable to apply patches; patch is not installed')
            return False

        for patch in patches:
            print('({})'.format(os.path.basename(patch)))

            if not PATCH.execute([
                    '--batch',
                    '--forward',
                    '--ignore-whitespace',
                    '--input={}'.format(patch),
                    '--strip=1',
                    ], cwd=build_dir):
                err('failed to apply patch')
                return False

    return True



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

from collections import defaultdict
from releng_tool.util.io import ensure_dir_exists
from releng_tool.util.log import debug
from releng_tool.util.log import verbose
import math
import os
import pickle

try:
    from time import clock as capture_clock
except ImportError:
    from time import monotonic as capture_clock

# optional imports
try:
    # disable xwindows backend (as it is not required and may cause issue with
    # systems without a display configured)
    import matplotlib
    matplotlib.use('Agg')

    import matplotlib.pyplot as plt
    import numpy
    has_matplotlib = True
except ImportError:
    has_matplotlib = False

# filename for the statistics database
STATISTICS_NAME = 'statistics.dat'


class RelengStats:
    """
    statistics tracking

    Registry provides a means for external implementation to hook into various
    stages of a release engineering process.

    Args:
        opts: options used to configure the engine

    Attributes:
        cache: cache of statistics for this runtime
        dat_file: file to store persisted statistics
        data: dictionary of data that can be persisted
        opts: options used to configure the engine
        out_dir: directory to generate final statistics to
    """
    def __init__(self, opts):
        self.cache = defaultdict(lambda: defaultdict(dict))
        self.opts = opts
        self.out_dir = os.path.join(self.opts.out_dir, 'misc')
        self.dat_file = os.path.join(self.out_dir, STATISTICS_NAME)

        self.data = {}

    def load(self):
        """
        load any persisted statistics

        Will load any statistics which may have been persisted from a previous
        run. This is to help render a "complete" report of statistics when
        re-running releng-tool with packages which may already been completed.
        """
        if not os.path.exists(self.dat_file):
            return

        try:
            with open(self.dat_file, 'rb') as f:
                self.data = pickle.load(f)
            debug('loaded statistics')
        except IOError:
            verbose('failed to load original statistics (io error)')
        except ValueError:
            verbose('failed to load original statistics (pickle error)')

    def save(self, desc=None):
        """
        save statistics for future reference

        Will save any statistics which should be persisted for future
        considerations. This is to help render a "complete" report of statistics
        when re-running releng-tool with packages which may already been
        completed.

        Args:
            desc (optional): description of this save event (for logging)
        """

        if not ensure_dir_exists(self.out_dir):
            verbose('unable to generate output directory for statistics')
            return None

        if desc:
            desc = ' ({})'.format(desc)
        else:
            desc = ''

        try:
            with open(self.dat_file, 'wb') as f:
                pickle.dump(self.data, f, protocol=2) # 2 for py2/py3 support
            debug('saved statistics' + desc)
        except IOError:
            verbose('failed to save statistics' + desc)

    def track_duration_start(self, pkg, stage):
        """
        track a duration start

        To be invoked when tracking the start of a package event for a given
        stage. This call is to be used with ``track_duration_end``, to help
        track the duration of a package's stage.

        Args:
            pkg: the package
            stage: the stage which has started
        """
        self.cache[pkg][stage]['start'] = capture_clock()

    def track_duration_end(self, pkg, stage, save=True):
        """
        track a duration end

        To be invoked when tracking the end of a package event for a given
        stage. This call is to be used with ``track_duration_start``, to help
        track the duration of a package's stage.

        Args:
            pkg: the package
            stage: the stage which has ended
            save (optional): automatically save the duration (default: True)
        """
        end_time = capture_clock()
        start_time = self.cache[pkg][stage]['start']

        if 'duration' not in self.data:
            self.data['duration'] = {}
        if pkg not in self.data['duration']:
            self.data['duration'][pkg] = {}
        if stage not in self.data['duration'][pkg]:
            self.data['duration'][pkg][stage] = {}

        self.data['duration'][pkg][stage] = end_time - start_time

        if save:
            self.save(desc='{}-{}'.format(pkg, stage))

    def generate(self):
        """
        generate a final report of statistics

        To be invoked at the end of a releng-tool process, this call will
        generate reports/etc. for any tracked statistics information based on
        the current and previous invoked executions (if any).
        """
        if not ensure_dir_exists(self.out_dir):
            verbose('unable to generate output directory for statistics')
            return None

        self._generate_duration()

    def _generate_duration(self):
        """
        generate duration-related statistics

        When generating a statistics report, this call creating/adds information
        about durations which may have been captured.
        """

        if 'duration' not in self.data:
            return

        durations = self.data['duration']

        pkgs = list(durations.keys())
        pkgs = sorted(pkgs)

        categories = set()
        for pkg_data in durations.values():
            categories.update(pkg_data.keys())
        categories = sorted(categories)

        ordered_categories = [
            'boot',
            'fetch',
            'extract',
            'patch',
            'configure',
            'build',
            'install',
            'post',
        ]

        for ordered_category in ordered_categories:
            if ordered_category not in categories:
                ordered_categories.remove(ordered_category)
        for category in categories:
            if category not in ordered_categories:
                ordered_categories.append(category)
        categories = ordered_categories

        # duration statistics to csv
        verbose('generating duration statistics (csv)...')
        dur_csv = os.path.join(self.out_dir, 'durations.csv')
        try:
            with open(dur_csv, 'w') as f:
                # header
                f.write('# pkg')
                for category in categories:
                    f.write(',' + category)
                f.write('\n')

                # data
                for pkg in pkgs:
                    f.write(pkg)

                    for category in categories:
                        if category in durations[pkg]:
                            value = durations[pkg][category]
                        else:
                            value = 0
                        f.write(',' + str(int(value)))
                    f.write('\n')
        except IOError as e:
            verbose('failed to write duration statistics: {}', e)

        # duration statistics to plot (if available)
        if has_matplotlib:
            verbose('generating duration statistics (pdf)...')

            BAR_HEIGHT = 0.4
            EXTRA_HEIGHT = 1
            FIG_WIDTH = 10
            fig_height_pkgs = (BAR_HEIGHT + EXTRA_HEIGHT) * len(pkgs)
            fig_height_total = (BAR_HEIGHT + EXTRA_HEIGHT) * (len(pkgs) + 1)

            figsize_pkgs = (FIG_WIDTH, fig_height_pkgs)
            figsize_total = (FIG_WIDTH, fig_height_total)

            fig_pkgs, ax_pkgs = plt.subplots(figsize=figsize_pkgs)
            fig_total, ax_total = plt.subplots(figsize=figsize_total)
            axs = [ax_pkgs, ax_total]
            figs = [fig_pkgs, fig_total]

            pkgs.reverse()
            pkgs_total = list(pkgs)
            pkgs_total.insert(0, 'total')

            offset = [0] * len(pkgs)
            offset_total = [0] * len(pkgs_total)
            for category in categories:
                width = []
                width_total = []
                total = 0

                for pkg in pkgs:
                    if category in durations[pkg]:
                        duration = durations[pkg][category]
                        width.append(duration)
                        width_total.append(duration)
                        total += duration
                    else:
                        width.append(0)
                        width_total.append(0)
                width_total.insert(0, total)

                ax_pkgs.barh(pkgs, width, height=BAR_HEIGHT,
                    left=offset, label=category)
                ax_total.barh(pkgs_total, width_total, height=BAR_HEIGHT,
                    left=offset_total, label=category)
                offset = numpy.add(offset, width)
                offset_total = numpy.add(offset_total, width_total)

            # provide some spacing near the right
            xlim = int(math.ceil(max(offset) / 10.)) * 10
            if xlim - max(offset) < 10:
                xlim += 10
            ax_pkgs.set_xlim([0, xlim])

            xlim_total = int(math.ceil(max(offset_total) / 10.)) * 10
            if xlim_total - max(offset_total) < 10:
                xlim_total += 10
            ax_total.set_xlim([0, xlim_total])

            # labels
            for ax in axs:
                ax.set_title('Package Stage Durations')
                ax.set_xlabel('Duration (seconds)')
                ax.legend()
                ax.grid(axis='x', linestyle=':', linewidth=0.4)

            # ensure rotated labels state in render area
            for fig in figs:
                fig.tight_layout()

            # generate figures
            dur_pdf = os.path.join(self.out_dir, 'durations.pdf')
            fig_pkgs.savefig(dur_pdf)

            dur_pdf_total = os.path.join(self.out_dir, 'durations-total.pdf')
            fig_total.savefig(dur_pdf_total)

            # close/cleanup figures
            plt.close()
        else:
            debug('duration statistics plot not supported')



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from io import open
from releng_tool.tool import RelengTool
from releng_tool.util.log import err

try:
    import configparser
except ImportError:
    import ConfigParser as configparser

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

#: executable used to run git commands
GIT_COMMAND = 'git'

#: list of environment keys to filter from a environment dictionary
GIT_SANITIZE_ENV_KEYS = [
    # disable repository location overrides
    'GIT_ALTERNATE_OBJECT_DIRECTORIES',
    'GIT_DIR',
    'GIT_INDEX_FILE',
    'GIT_OBJECT_DIRECTORY',
    'GIT_WORK_TREE',
    # remove the possibility for authenticated prompts
    'GIT_ASKPASS',
    'SSH_ASKPASS',
    # misc
    'GIT_FLUSH',
]

#: dictionary of environment entries append to the environment dictionary
GIT_EXTEND_ENV = {
    # prevent the terminal prompt from being shown
    'GIT_TERMINAL_PROMPT': '0',
}

class GitTool(RelengTool):
    """
    git host tool

    Provides addition helper methods for git-based tool interaction.
    """

    def extract_submodule_revision(self, bare_dir):
        """
        extract a submodule revision

        Attempts to extract the HEAD reference of a submodule based off a
        provided bare Git repository. This is to help support processing Git
        submodules which do not have a branch/version explicitly set for module,
        which is required for (at least) recursive submodule processing.

        Args:
            bare_dir: the bare repository

        Returns:
            the revision; ``None`` when a revision cannot be extracted
        """

        rv, ref = self.execute_rv('--git-dir=' + bare_dir, 'show-ref', '--head')
        if rv != 0:
            err('failed to extract a submodule revision')
            return None

        # a `--head` fetch may fetch more than one reference; extract the first
        # entry and remove any known ref prefix from it
        revision = ref.split(None, 2)[1]
        if revision.startswith('refs/heads/'):
            revision = revision[len('refs/heads/'):]
        elif revision.startswith('refs/remotes/origin/'):
            revision = revision[len('refs/remotes/origin/'):]
        return revision

    def parse_cfg_file(self, target):
        """
        return a configuration parser for a provided git configuration file

        Returns a prepared configuration parser based of a Git configuration
        file provided. In the event that the file cannot be parsed, this call
        will return a ``None`` value.

        Args:
            target: the file to parse

        Returns:
            the parser; ``None`` when a parsing error is detected
        """

        with open(target, mode='r', encoding='utf_8') as f:
            data = '\n'.join(f.readlines())

        return self.parse_cfg_str(data)

    def parse_cfg_str(self, value):
        """
        return a configuration parser for a provided git configuration string

        Returns a prepared configuration parser based of a Git configuration
        value provided. In the event that the value cannot be parsed, this call
        will return a ``None`` value.

        Args:
            value: the value to parse

        Returns:
            the parser; ``None`` when a parsing error is detected
        """

        cfg = configparser.ConfigParser(allow_no_value=True)
        try:
            if hasattr(cfg, 'read_string'):
                cfg.read_string(value)
            else:
                # strip whitespaces from lines for python 2.7
                value = '\n'.join([line.strip() for line in value.splitlines()])

                fp = StringIO(value)
                cfg.readfp(fp)  # pylint: disable=W1505
        except configparser.Error:
            return None

        return cfg

#: git host tool helper
GIT = GitTool(GIT_COMMAND,
    env_sanitize=GIT_SANITIZE_ENV_KEYS, env_include=GIT_EXTEND_ENV)



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

from releng_tool.exceptions import RelengToolMissingConfigurationError
from releng_tool.exceptions import RelengToolMissingPackagesError
from tests import prepare_testenv
from tests import run_testenv
import os
import unittest

class TestEngineRunDefaults(unittest.TestCase):
    def test_engine_run_defaults_dirs(self):
        with prepare_testenv(template='minimal') as engine:
            # check if root directory is tracked
            root_dir = engine.opts.root_dir
            self.assertTrue(os.path.exists(root_dir))

            # ensure default paths for other directories are populated
            expected_cache_dir = os.path.join(root_dir, 'cache')
            expected_dl_dir = os.path.join(root_dir, 'dl')
            expected_out_dir = os.path.join(root_dir, 'output')
            self.assertEqual(engine.opts.cache_dir, expected_cache_dir)
            self.assertEqual(engine.opts.dl_dir, expected_dl_dir)
            self.assertEqual(engine.opts.out_dir, expected_out_dir)

    def test_engine_run_defaults_noconfig(self):
        with self.assertRaises(RelengToolMissingConfigurationError):
            run_testenv()

    def test_engine_run_defaults_nopackages(self):
        with self.assertRaises(RelengToolMissingPackagesError):
            run_testenv(template='no-packages')

    def test_engine_run_defaults_options_jobs(self):
        with prepare_testenv() as engine:
            self.assertNotEqual(engine.opts.jobs, 0)
            self.assertEqual(engine.opts.jobsconf, 0)

    def test_engine_run_defaults_options_logging(self):
        with prepare_testenv() as engine:
            self.assertFalse(engine.opts.debug)
            self.assertFalse(engine.opts.no_color_out)
            self.assertFalse(engine.opts.verbose)

    def test_engine_run_defaults_options_modes(self):
        with prepare_testenv() as engine:
            self.assertFalse(engine.opts.devmode)
            self.assertFalse(engine.opts.force)
            self.assertFalse(engine.opts.local_srcs)



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.tool import RelengTool
from releng_tool.util.io import execute
from releng_tool.util.log import debug
from releng_tool.util.win32 import find_win32_python_interpreter
import os
import sys

#: executable used to run python commands
PYTHON_COMMAND = 'python'

#: dictionary of environment entries append to the environment dictionary
PYTHON_EXTEND_ENV = {
    # prevent user site-packages from being included
    'PYTHONNOUSERSITE': '1',
}

class PythonTool(RelengTool):
    """
    python host tool

    Provides addition helper methods for Python-based tool interaction.
    """

    def exists(self):
        """
        return whether or not the host tool exists

        Returns whether or not the tool is available on the host for use.

        Returns:
            ``True``, if the tool exists; ``False`` otherwise
        """
        if self.tool in RelengTool.detected:
            return RelengTool.detected[self.tool]

        found = True
        tool = self.tool

        if execute([tool] + self.exists_args, quiet=True, critical=False):
            found = True
        # if windows and a non-path entry, try to find the interpreter on the
        # local system
        elif sys.platform == 'win32' and os.path.basename(tool) == tool:
            debug('{} tool not available in path; '
                  'attempting to search the system...', tool)
            alt_tool = find_win32_python_interpreter(tool)
            if alt_tool:
                debug('{} tool to be replaced by: {}', tool, alt_tool)

                if execute([alt_tool] + self.exists_args, quiet=True,
                        critical=False):
                    found = True

                    # adjust the tool for this instance to the newly detected
                    # interpreter path
                    tool = alt_tool
                    self.tool = tool

        if found:
            debug('{} tool is detected on this system', tool)
            RelengTool.detected[tool] = True
        else:
            debug('{} tool is not detected on this system', tool)
            RelengTool.detected[tool] = False

        return RelengTool.detected[tool]

    def path(self, sysroot=None, prefix=None):
        """
        return a python path value for the python interpreter

        Returns the expected Python path (``PYTHONPATH``) value for the Python
        interpreter (without a prefix or system root container).

        Args:
            sysroot (optional): system root value to add
            prefix (optional): prefix value to add

        Returns:
            the python path
        """

        if not sysroot:
            sysroot = ''

        if not prefix:
            prefix = ''

        # determine interpreter's major/minor version for determined path
        if not hasattr(self, '_version_cache'):
            output = []
            version = ''
            if self.execute(['-c', "import sys; " +
                    "print('.'.join(map(str, sys.version_info[:2])))"],
                    capture=output, quiet=True):
                version = ''.join(output)
                if sys.platform == 'win32':
                    version = version.replace('.', '')
            self._version_cache = version

        if sys.platform != 'win32':
            base_path = os.path.join(sysroot + prefix,
                'lib', 'python' + self._version_cache)
        else:
            base_path = os.path.join(sysroot,
                'Python' + self._version_cache, 'Lib')
        return os.path.join(base_path, 'site-packages')

#: python host tool helper
PYTHON = PythonTool(PYTHON_COMMAND, env_include=PYTHON_EXTEND_ENV)



# -*- coding: utf-8 -*-
# Copyright 2021-2022 releng-tool

from tests import prepare_testenv
from tests import run_testenv
import os
import unittest

class TestEngineRunScripts(unittest.TestCase):
    def test_engine_run_scripts_include(self):
        with prepare_testenv(template='scripts-include') as engine:
            rv = engine.run()
            self.assertTrue(rv)

            file_flags = [
                os.path.join(engine.opts.target_dir, 'invoked-stage1'),
                os.path.join(engine.opts.target_dir, 'invoked-stage2'),
            ]

            for file_flag in file_flags:
                self.assertTrue(os.path.exists(file_flag))

    def test_engine_run_scripts_invalid_bootstrap(self):
        rv = run_testenv(template='scripts-invalid-bootstrap')
        self.assertFalse(rv)

    def test_engine_run_scripts_invalid_build(self):
        rv = run_testenv(template='scripts-invalid-build')
        self.assertFalse(rv)

    def test_engine_run_scripts_invalid_configure(self):
        rv = run_testenv(template='scripts-invalid-configure')
        self.assertFalse(rv)

    def test_engine_run_scripts_invalid_install(self):
        rv = run_testenv(template='scripts-invalid-install')
        self.assertFalse(rv)

    def test_engine_run_scripts_invalid_post(self):
        rv = run_testenv(template='scripts-invalid-post')
        self.assertFalse(rv)

    def test_engine_run_scripts_valid(self):
        with prepare_testenv(template='scripts-valid') as engine:
            rv = engine.run()
            self.assertTrue(rv)

            file_flags = [
                os.path.join(engine.opts.target_dir, 'invoked-bootstrap'),
                os.path.join(engine.opts.target_dir, 'invoked-configure'),
                os.path.join(engine.opts.target_dir, 'invoked-build'),
                os.path.join(engine.opts.target_dir, 'invoked-install'),
                os.path.join(engine.opts.target_dir, 'invoked-post'),
            ]

            for file_flag in file_flags:
                self.assertTrue(os.path.exists(file_flag))



# -*- coding: utf-8 -*-
# Copyright 2022 releng-tool

from releng_tool import __version__ as releng_version
from releng_tool.support import require_version
import unittest


class TestSupport(unittest.TestCase):
    def test_support_require_version(self):
        # existing version should report has valid
        result = require_version(releng_version)
        self.assertTrue(result)

        # an old version should report has valid
        result = require_version('0.0')
        self.assertTrue(result)

        # fail version check on a future version which does not exist
        result = require_version('999999', critical=False)
        self.assertFalse(result)

        # fail version check on a future version which does not exist (critical)
        with self.assertRaises(SystemExit):
            require_version('999999')



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

from releng_tool.packages.exceptions import RelengToolInvalidPackageKeyValue
from tests.support.pkg_config_test import TestPkgConfigsBase

class TestPkgConfigsGit(TestPkgConfigsBase):
    def test_pkgconfig_git_config_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('git-config-invalid-base-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('git-config-invalid-key-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('git-config-invalid-strs')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('git-config-invalid-value-type')

    def test_pkgconfig_git_config_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.git_config)

        pkg, _, _ = self.LOAD('conf-defs-valid')
        self.assertIsNone(pkg.git_config)

        pkg, _, _ = self.LOAD('install-defs-valid')
        self.assertIsNone(pkg.git_config)

    def test_pkgconfig_git_config_valid(self):
        pkg, _, _ = self.LOAD('git-config-valid')
        self.assertDictEqual(pkg.git_config, {
            'key1': 'val1',
            'key2': None,
            'key3': 'val3',
        })

    def test_pkgconfig_git_depth_invalid_type(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('git-depth-invalid-type')

    def test_pkgconfig_git_depth_invalid_value(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('git-depth-invalid-value')

    def test_pkgconfig_git_depth_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.git_depth)

    def test_pkgconfig_git_depth_valid_nonzero(self):
        pkg, _, _ = self.LOAD('git-depth-valid-nonzero')
        self.assertEqual(pkg.git_depth, 50)

    def test_pkgconfig_git_depth_valid_zero(self):
        pkg, _, _ = self.LOAD('git-depth-valid-zero')
        self.assertEqual(pkg.git_depth, 0)

    def test_pkgconfig_git_refspecs_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('git-refspecs-invalid-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('git-refspecs-invalid-value')

    def test_pkgconfig_git_refspecs_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.git_refspecs)

    def test_pkgconfig_git_refspecs_valid(self):
        pkg, _, _ = self.LOAD('git-refspecs-valid-empty')
        self.assertListEqual(pkg.git_refspecs, [])

        pkg, _, _ = self.LOAD('git-refspecs-valid-multiple')
        self.assertListEqual(pkg.git_refspecs, [
            'refspec1',
            'refspec2',
            'refspec3',
        ])

        pkg, _, _ = self.LOAD('git-refspecs-valid-single')
        self.assertListEqual(pkg.git_refspecs, [
            'refspec',
        ])

    def test_pkgconfig_git_submodules_disabled(self):
        pkg, _, _ = self.LOAD('git-submodules-disabled')
        self.assertFalse(pkg.git_submodules)

    def test_pkgconfig_git_submodules_enabled(self):
        pkg, _, _ = self.LOAD('git-submodules-enabled')
        self.assertTrue(pkg.git_submodules)

    def test_pkgconfig_git_submodules_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('git-submodules-invalid')

    def test_pkgconfig_git_submodules_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.git_submodules)

    def test_pkgconfig_git_verify_disabled(self):
        pkg, _, _ = self.LOAD('git-verify-disabled')
        self.assertFalse(pkg.git_verify_revision)

    def test_pkgconfig_git_verify_enabled(self):
        pkg, _, _ = self.LOAD('git-verify-enabled')
        self.assertTrue(pkg.git_verify_revision)

    def test_pkgconfig_git_verify_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('git-verify-invalid')

    def test_pkgconfig_git_verify_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.git_verify_revision)



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

from releng_tool.packages.exceptions import RelengToolInvalidPackageKeyValue
from tests.support.pkg_config_test import TestPkgConfigsBase

class TestPkgConfigsAutotools(TestPkgConfigsBase):
    def test_pkgconfig_autotools_autoreconf_disabled(self):
        pkg, _, _ = self.LOAD('autotools-autoreconf-disabled')
        self.assertFalse(pkg.autotools_autoreconf)

    def test_pkgconfig_autotools_autoreconf_enabled(self):
        pkg, _, _ = self.LOAD('autotools-autoreconf-enabled')
        self.assertTrue(pkg.autotools_autoreconf)

    def test_pkgconfig_autotools_autoreconf_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('autotools-autoreconf-invalid')

    def test_pkgconfig_autotools_autoreconf_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.autotools_autoreconf)



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

from functools import partial
import sys

# import compatible `copytree` calls, use shutil's copytree with dirs_exist_ok
# set; otherwise fallback to distutil's copy_tree call
if sys.version_info >= (3, 8):
    from shutil import copytree as shutil_copy_tree

    _CompatCopyTreeError = IOError
    _compat_copy_tree = partial(shutil_copy_tree, dirs_exist_ok=True)
else:
    from distutils.dir_util import DistutilsFileError
    from distutils.dir_util import copy_tree as distutils_copy_tree

    _CompatCopyTreeError = DistutilsFileError
    _compat_copy_tree = distutils_copy_tree

# import compatible `which` calls, use shutil's which if possible; otherwise
# fallback to distutil's find_executable call
if sys.version_info >= (3, 3):
    from shutil import which as shutil_which
    _compat_which = shutil_which
else:
    from distutils.spawn import find_executable as distutils_which
    _compat_which = distutils_which


# ######################################################################
# various python compatible classes/functions


class CopyTreeError(Exception):
    pass


def copy_tree(src, dst):
    try:
        _compat_copy_tree(src, dst)
    except _CompatCopyTreeError as e:
        raise CopyTreeError(str(e))


which = _compat_which



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

from releng_tool.tool import RelengTool
from releng_tool.util.log import log
from releng_tool.util.log import verbose

#: executable used to run gpg commands
GPG_COMMAND = 'gpg'

class GpgTool(RelengTool):
    """
    gpg host tool

    Provides addition helper methods for gpg-based tool interaction.
    """

    def validate(self, asc, target):
        """
        validate ascii-armored file against a target

        Accepting an ASCII-armored file, use gpg to validate the public key
        against the provided target.

        Args:
            asc: the asc file
            target: the target file

        Returns:
            ``True`` if the target has been validated; ``False`` otherwise
        """

        rv, out = self.execute_rv('--verify', asc, target)
        if rv == 0:
            verbose('validated: {}', asc)
        elif out:
            log(out)

        return rv == 0

#: gpg host tool helper
GPG = GpgTool(GPG_COMMAND)



# -*- coding: utf-8 -*-
# Copyright 2018-2022 releng-tool

from collections import OrderedDict
from releng_tool.util.io import ensure_dir_exists
from releng_tool.util.io import execute
from releng_tool.util.io import interpret_stem_extension as ise
from releng_tool.util.io import opt_file
from releng_tool.util.io import path_copy
from releng_tool.util.io import path_exists
from releng_tool.util.io import path_move
from releng_tool.util.io import path_remove
from releng_tool.util.io import prepare_arguments
from releng_tool.util.io import prepare_definitions
from releng_tool.util.io import prepend_shebang_interpreter as psi
from releng_tool.util.io import run_script
from releng_tool.util.io import touch
from releng_tool.util.log import is_verbose
from tests import compare_contents
from tests import prepare_workdir
from tests import redirect_stdout
import os
import sys
import unittest

ASSETS_DIR = 'assets'

class TestUtilIo(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        base_dir = os.path.dirname(os.path.realpath(__file__))
        cls.assets_dir = os.path.join(base_dir, ASSETS_DIR)

        def assertExists(cls, path, *args):
            cls.assertTrue(path_exists(path, *args),
                'missing file: ' + os.path.join(path, *args))
        cls.assertExists = assertExists

    def test_utilio_copy(self):
        check_dir_01 = os.path.join(self.assets_dir, 'copy-check-01')
        check_dir_02 = os.path.join(self.assets_dir, 'copy-check-02')

        with prepare_workdir() as work_dir:
            # (directories)

            # copy the first batch of assets into the working directory
            result = path_copy(check_dir_01, work_dir, critical=False)
            self.assertTrue(result)
            self.assertExists(work_dir, 'test-file-a')
            self.assertExists(work_dir, 'test-file-b')
            self.assertExists(work_dir, 'test-file-x')
            cc1 = compare_contents(
                os.path.join(check_dir_01, 'test-file-b'),
                os.path.join(work_dir, 'test-file-b'))
            self.assertIsNone(cc1, cc1)

            # override the working directory with the second batch
            result = path_copy(check_dir_02, work_dir, critical=False)
            self.assertTrue(result)
            self.assertExists(work_dir, 'test-file-a')
            self.assertExists(work_dir, 'test-file-b')
            self.assertExists(work_dir, 'test-file-c')
            self.assertExists(work_dir, 'test-file-x')
            cc2 = compare_contents(
                os.path.join(check_dir_02, 'test-file-b'),
                os.path.join(work_dir, 'test-file-b'))
            self.assertIsNone(cc2, cc2)

            # (files)
            sub_dir = os.path.join(work_dir, 'sub', 'dir', 'xyz', '')
            copied_file_a = os.path.join(work_dir, 'test-file-a')
            copied_file_b = os.path.join(work_dir, 'test-file-b')
            target_ow = os.path.join(sub_dir, 'test-file-a')

            # copy individual files (file to new directory)
            result = path_copy(copied_file_a, sub_dir, critical=False)
            self.assertTrue(result)
            self.assertExists(target_ow)
            cc3 = compare_contents(copied_file_a, target_ow)
            self.assertIsNone(cc3, cc3)

            # copy individual files (overwrite)
            result = path_copy(copied_file_b, target_ow, critical=False)
            self.assertTrue(result)
            cc4 = compare_contents(copied_file_b, target_ow)
            self.assertIsNone(cc4, cc4)

            # attempt to copy a missing file
            missing_file = os.path.join(work_dir, 'test-file-missing')
            target = os.path.join(work_dir, 'container')
            result = path_copy(missing_file, target, critical=False)
            self.assertFalse(result)

            # attempt to copy a missing file (critical)
            with self.assertRaises(SystemExit):
                path_copy(missing_file, target)

            # attempt to copy a file to itself
            src = copied_file_a
            with self.assertRaises(SystemExit):
                path_copy(src, src)

            # attempt to copy a directory to itself in legacy python, to ensure
            # `DistutilsFileError` is handled properly (windows only)
            if sys.version_info < (3, 0) and sys.platform == 'win32':
                with self.assertRaises(SystemExit):
                    path_copy(work_dir, work_dir)

            # force a directory target with a non-trailing path separator
            force_src = os.path.join(work_dir, 'test-file-a')
            self.assertExists(force_src)

            force1 = os.path.join(work_dir, 'force1')
            result = path_copy(force_src, force1, critical=False)
            self.assertTrue(result)
            self.assertTrue(os.path.isfile(force1))

            force2 = os.path.join(work_dir, 'force2')
            result = path_copy(force_src, force2, critical=False, dst_dir=True)
            self.assertTrue(result)
            self.assertTrue(os.path.isdir(force2))

    def test_utilio_ensuredirexists(self):
        with prepare_workdir() as work_dir:
            result = ensure_dir_exists(work_dir)
            self.assertTrue(result)

            new_dir = os.path.join(work_dir, 'test1')
            self.assertFalse(os.path.exists(new_dir))

            result = ensure_dir_exists(new_dir)
            self.assertTrue(result)
            self.assertTrue(os.path.exists(new_dir))

            new_file = os.path.join(work_dir, 'test2')
            with open(new_file, 'ab'):
                pass
            self.assertTrue(os.path.isfile(new_file))

            result = ensure_dir_exists(new_file)
            self.assertFalse(result)
            self.assertTrue(os.path.isfile(new_file))

    def test_utilio_execution(self):
        result = execute(None, quiet=True, critical=False)
        self.assertFalse(result)

        result = execute([], quiet=True, critical=False)
        self.assertFalse(result)

        test_cmd = ['python', '-c', 'print("Hello")']
        result = execute(test_cmd, quiet=True, critical=False)
        self.assertTrue(result)

        result = execute(['an_unknown_command'], quiet=True, critical=False)
        self.assertFalse(result)

        # skip output checks if verbose mode is enabled
        if is_verbose():
            raise unittest.SkipTest(
                'ignoring execution output checks while in verbose mode')

        # verify output
        with redirect_stdout() as stream:
            test_cmd = [sys.executable, '-c', 'print("Hello")']
            result = execute(test_cmd, critical=False)
            self.assertTrue(result)
        self.assertEqual(stream.getvalue().strip(), 'Hello')

        # verify quiet mode
        with redirect_stdout() as stream:
            test_cmd = [sys.executable, '-c', 'print("Hello")']
            result = execute(test_cmd, quiet=True, critical=False)
            self.assertTrue(result)
        self.assertEqual(stream.getvalue().strip(), '')

        # verify capture mode which will be silent
        out = []
        with redirect_stdout() as stream:
            test_cmd = [sys.executable, '-c', 'print("Hello")']
            result = execute(test_cmd, critical=False, capture=out)
            self.assertTrue(result)
        self.assertEqual(''.join(out), 'Hello')
        self.assertEqual(stream.getvalue().strip(), '')

        # verify capture mode that is also verbose
        out = []
        with redirect_stdout() as stream:
            test_cmd = [sys.executable, '-c', 'print("Hello")']
            result = execute(test_cmd, quiet=False, critical=False, capture=out)
            self.assertTrue(result)
        self.assertEqual(''.join(out), 'Hello')
        self.assertEqual(stream.getvalue().strip(), 'Hello')

    def test_utilio_ise(self):
        provided = 'my-file.txt'
        expected = ('my-file', 'txt')
        self.assertEqual(ise(provided), expected)

        provided = 'my-file.tar.Z'
        expected = ('my-file', 'tar.Z')
        self.assertEqual(ise(provided), expected)

        provided = 'my-file.tar.gz'
        expected = ('my-file', 'tar.gz')
        self.assertEqual(ise(provided), expected)

        provided = 'my-file.tar.xz'
        expected = ('my-file', 'tar.xz')
        self.assertEqual(ise(provided), expected)

        provided = 'my.file.name.dat'
        expected = ('my.file.name', 'dat')
        self.assertEqual(ise(provided), expected)

        provided = 'my-file'
        expected = ('my-file', None)
        self.assertEqual(ise(provided), expected)

        provided = None
        expected = (None, None)
        self.assertEqual(ise(provided), expected)

    def test_utilio_move(self):
        with prepare_workdir() as work_dir:
            def _(*args):
                return os.path.join(work_dir, *args)

            # setup
            directories = [
                _('dir1'),
                _('dir2', 'dir3'),
                _('dir4', 'dir5', 'dir6'),
            ]
            for dir_ in directories:
                os.makedirs(dir_)

            files = [
                _('file1'),
                _('dir2', 'file2'),
                _('dir2', 'dir3', 'file3'),
                _('dir4', 'file4'),
                _('dir4', 'dir5', 'file5'),
                _('dir4', 'dir5', 'dir6', 'file6'),
            ]
            for file in files:
                with open(file, 'a') as f:
                    f.write(file)

            # (simple file move)
            src = _('file1')
            dst = _('file7')
            moved = path_move(src, dst, critical=False)
            self.assertTrue(moved)
            self.assertFalse(os.path.isfile(src))
            self.assertTrue(os.path.isfile(dst))

            # (another file move)
            src = _('dir4', 'dir5', 'dir6', 'file6')
            dst = _('dir9', 'dir10', 'file8')
            moved = path_move(src, dst, critical=False)
            self.assertTrue(moved)
            self.assertFalse(os.path.isfile(src))
            self.assertTrue(os.path.isfile(dst))

            # (another file move)
            src = _('dir9', 'dir10', 'file8')
            dst = _('dir9', 'dir11', '')
            moved = path_move(src, dst, critical=False)
            self.assertTrue(moved)
            self.assertFalse(os.path.isfile(src))
            self.assertTrue(os.path.isfile(_('dir9', 'dir11', 'file8')))

            # (overwriting file move)
            src = _('dir2', 'file2')
            dst = _('file7')
            moved = path_move(src, dst, critical=False)
            self.assertTrue(moved)
            self.assertFalse(os.path.isfile(src))
            self.assertTrue(os.path.isfile(dst))

            # (bad file move attempt)
            src = _('file7')
            dst_part = _('dir4', 'file4')
            dst = _(dst_part, 'bad')
            moved = path_move(src, dst, quiet=True, critical=False)
            self.assertFalse(moved)
            self.assertTrue(os.path.isfile(src))
            self.assertTrue(os.path.isfile(dst_part))

            # (bad directory move self container)
            src = _('dir2')
            dst = _('dir2', 'dir3')
            moved = path_move(src, dst, quiet=True, critical=False)
            self.assertFalse(moved)
            self.assertTrue(os.path.isdir(src))
            self.assertTrue(os.path.isdir(dst))

            # (simple directory move)
            src = _('dir2', 'dir3')
            dst = _('dir4')
            moved = path_move(src, dst, critical=False)
            self.assertTrue(moved)
            self.assertFalse(os.path.isdir(src))
            self.assertFalse(os.path.isdir(_(dst, 'dir3')))
            self.assertTrue(os.path.isfile(_(dst, 'file3')))

            # (another directory move)
            src = _('dir4')
            dst = _('dir9', 'dir10')
            moved = path_move(src, dst, critical=False)
            self.assertTrue(moved)
            self.assertFalse(os.path.isdir(src))
            self.assertTrue(os.path.isdir(dst))
            self.assertTrue(os.path.isfile(_(dst, 'file3')))
            self.assertTrue(os.path.isfile(_(dst, 'file4')))
            self.assertTrue(os.path.isdir(_(dst, 'dir5')))

            # (check directory replacing a file)
            src = _('dir9')
            dst = _('file7')
            self.assertTrue(os.path.isdir(src))
            self.assertTrue(os.path.isfile(dst))
            moved = path_move(src, dst, quiet=True, critical=False)
            self.assertTrue(moved)
            self.assertFalse(os.path.isdir(src))
            self.assertTrue(os.path.isdir(dst))

    def test_utilio_optfile(self):
        with prepare_workdir() as work_dir:
            def _(*args):
                return os.path.join(work_dir, *args)

            # setup
            files = [
                _('file1'),
                _('file2.py'),
                _('file3'),
                _('file3.py'),
            ]
            for file in files:
                with open(file, 'a') as f:
                    f.write(file)

            # checks
            src = _('file1')
            target, existence = opt_file(src)
            self.assertTrue(existence)
            self.assertEqual(target, src)

            src = _('file2')
            opt = _('file2.py')
            target, existence = opt_file(src)
            self.assertTrue(existence)
            self.assertEqual(target, opt)

            src = _('file3')
            target, existence = opt_file(src)
            self.assertTrue(existence)
            self.assertEqual(target, src)

            src = _('file4')
            target, existence = opt_file(src)
            self.assertFalse(existence)
            self.assertEqual(target, src)

    def test_utilio_prepare_helpers(self):
        prepared = prepare_arguments(None)
        expected = []
        self.assertEqual(prepared, expected)

        prepared = prepare_arguments({})
        expected = []
        self.assertEqual(prepared, expected)

        args = OrderedDict()
        args['foo'] = 'bar'
        args['xyz'] = ''
        prepared = prepare_arguments(args)
        expected = ['foo', 'bar', 'xyz']
        self.assertEqual(prepared, expected)

        args = OrderedDict()
        args['foo'] = 'bar'
        args['test'] = None
        prepared = prepare_arguments(args)
        expected = ['foo', 'bar']
        self.assertEqual(prepared, expected)

        prepared = prepare_definitions(None)
        expected = []
        self.assertEqual(prepared, expected)

        prepared = prepare_definitions({})
        expected = []
        self.assertEqual(prepared, expected)

        args = OrderedDict()
        args['foo'] = 'bar'
        args['xyz'] = ''
        prepared = prepare_definitions(args)
        expected = ['foo=bar', 'xyz']
        self.assertEqual(prepared, expected)

        args = OrderedDict()
        args['foo'] = 'bar'
        args['test'] = None
        prepared = prepare_definitions(args)
        expected = ['foo=bar']
        self.assertEqual(prepared, expected)

    def test_utilio_remove(self):
        with prepare_workdir() as work_dir:
            def _(*args):
                return os.path.join(work_dir, *args)

            # setup
            directories = [
                _('dir1'),
                _('dir2', 'dir3'),
                _('dir4', 'dir5', 'dir6'),
            ]
            for dir_ in directories:
                os.makedirs(dir_)

            files = [
                _('file1'),
                _('dir2', 'file2'),
                _('dir2', 'dir3', 'file3'),
                _('dir4', 'file4'),
                _('dir4', 'dir5', 'file5'),
                _('dir4', 'dir5', 'dir6', 'file6'),
            ]
            for file in files:
                with open(file, 'a') as f:
                    f.write(file)

            path = _('file7')
            removed = path_remove(path)
            self.assertTrue(removed)
            self.assertFalse(os.path.isfile(path))

            container = _('dir2')
            path = _(container, 'file2')
            removed = path_remove(path)
            self.assertTrue(removed)
            self.assertFalse(os.path.isfile(path))
            self.assertTrue(os.path.isdir(container))
            self.assertFalse(os.path.isfile(path))

            container = _('dir2')
            path = _(container, 'dir3')
            file = _(path, 'file3')
            removed = path_remove(path)
            self.assertTrue(removed)
            self.assertFalse(os.path.isdir(path))
            self.assertFalse(os.path.isfile(file))
            self.assertTrue(os.path.isdir(container))

            path = _('dir4')
            files = [
                _(path, 'file4'),
                _(path, 'dir5', 'file5'),
                _(path, 'dir5', 'dir6', 'file6'),
            ]
            removed = path_remove(path)
            self.assertTrue(removed)
            self.assertFalse(os.path.isdir(path))
            for file in files:
                self.assertFalse(os.path.isfile(file))

            # allow noop calls to be true
            path = _('missing')
            removed = path_remove(path)
            self.assertTrue(removed)

    def test_utilio_runscript(self):
        with prepare_workdir() as work_dir:
            valid_script = os.path.join(work_dir, 'valid')
            with open(valid_script, 'a') as f:
                f.write('test=1\n')

            result = run_script(valid_script, {})
            self.assertEqual(result['test'], 1)

            invalid_script = os.path.join(work_dir, 'invalid')
            with open(invalid_script, 'a') as f:
                f.write('bad-line\n')

            result = run_script(invalid_script, {})
            self.assertIsNone(result)

            with self.assertRaises(NameError):
                run_script(invalid_script, {}, catch=False)

    def test_utilio_shebang_interpreter(self):
        si_dir = os.path.join(self.assets_dir, 'shebang-interpreter')
        si01 = [os.path.join(si_dir, 'interpreter')]
        si02 = [os.path.join(si_dir, 'interpreter-arg')]
        si03 = [os.path.join(si_dir, 'interpreter-args-multiple')]
        si04 = [os.path.join(si_dir, 'interpreter-extremely-long')]
        si05 = [os.path.join(si_dir, 'interpreter-whitespaces')]
        si06 = [os.path.join(si_dir, 'example.py')]

        # encode helper as helper utility will execute with encoded strings
        def E(val):
            rv = []
            for v in val:
                rv.append(v.encode())
            return rv

        # simple interpreter
        self.assertEqual(psi(si01), [b'interpreter'] + E(si01))
        # interpreter with a single argument
        self.assertEqual(psi(si02), [b'interpreter', b'arg'] + E(si02))
        # interpreter with a single argument (with whitespaces)
        self.assertEqual(psi(si03), [b'interpreter', b'arg1 arg2'] + E(si03))
        # too long of an interpreter
        self.assertEqual(psi(si04), si04)
        # interpreter with whitespaces
        self.assertEqual(psi(si05), [b'interpreter'] + E(si05))
        # real example of an interpreter
        self.assertEqual(psi(si06), [b'/usr/bin/env', b'python'] + E(si06))

    def test_utilio_touch(self):
        with prepare_workdir() as work_dir:
            test_file = os.path.join(work_dir, 'test-file')

            created = touch(test_file)
            self.assertTrue(created)

            exists = os.path.isfile(test_file)
            self.assertTrue(exists)

            updated = touch(test_file)
            self.assertTrue(updated)



# -*- coding: utf-8 -*-
# Copyright 2022 releng-tool

from releng_tool.api import RelengExtractOptions
from releng_tool.extract.archive import extract
from tests import prepare_workdir
from tests.support import fetch_unittest_assets_dir
import json
import os
import posixpath
import unittest


class TestExtractArchive(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.assets_dir = fetch_unittest_assets_dir()
        cls.sample_files = os.path.join(cls.assets_dir, 'sample-files')

    def run(self, result=None):
        self.opts = RelengExtractOptions()
        self.opts.strip_count = 0

        with prepare_workdir() as work_dir:
            self.opts.work_dir = work_dir

            super(TestExtractArchive, self).run(result)

    def test_extract_archive_override_invalid(self):
        cache_file = os.path.join(self.opts.work_dir, 'mock.dat')
        self.opts.cache_file = cache_file

        self.opts._extract_override = {
            'dat': os.path.join(self.opts.work_dir, 'missing-cmd'),
        }

        extracted = extract(self.opts)
        self.assertFalse(extracted)

    def test_extract_archive_override_valid(self):
        cache_file = os.path.join(self.opts.work_dir, 'mock.dat')
        self.opts.cache_file = cache_file

        override_cmd = os.path.join(self.assets_dir, 'test-invoke.py')
        self.opts._extract_override = {
            'dat': 'python ' + override_cmd + ' {file} {dir}',
        }

        extracted = extract(self.opts)
        self.assertTrue(extracted)

        results = os.path.join(self.opts.work_dir, 'invoke-results.json')
        self.assertTrue(os.path.exists(results))

        with open(results, 'r') as f:
            data = json.load(f)

        self.assertTrue('args' in data)
        invoke_args = data['args']
        self.assertEqual(len(invoke_args), 3)
        self.assertEqual(invoke_args[0], override_cmd)
        self.assertEqual(invoke_args[1], cache_file)
        self.assertEqual(invoke_args[2], self.opts.work_dir)

    def test_extract_archive_tar_default(self):
        cache_file = os.path.join(self.sample_files, 'sample-files.tar')
        self.opts.cache_file = cache_file

        extracted = extract(self.opts)
        self.assertTrue(extracted)

        self._assertExtracted([
            'container/tar-file-container.txt',
            'container/',
            'tar-file-root',
        ])

    def test_extract_archive_tar_strip(self):
        cache_file = os.path.join(self.sample_files, 'sample-files.tgz')
        self.opts.cache_file = cache_file

        self.opts.strip_count = 1

        extracted = extract(self.opts)
        self.assertTrue(extracted)

        self._assertExtracted([
            'tgz-file-container.txt',
        ])

    def test_extract_archive_zip_default(self):
        cache_file = os.path.join(self.sample_files, 'sample-files.zip')
        self.opts.cache_file = cache_file

        extracted = extract(self.opts)
        self.assertTrue(extracted)

        self._assertExtracted([
            'container/zip-file-container.txt',
            'container/',
            'zip-file-root',
        ])

    def test_extract_archive_zip_strip(self):
        cache_file = os.path.join(self.sample_files, 'sample-files.zip')
        self.opts.cache_file = cache_file

        self.opts.strip_count = 1

        extracted = extract(self.opts)
        self.assertTrue(extracted)

        self._assertExtracted([
            'zip-file-container.txt',
        ])

    def _assertExtracted(self, required):
        entries = []
        for root, dirs, files in os.walk(self.opts.work_dir):
            for name in files:
                entries.append(os.path.join(root, name))

            for name in dirs:
                entries.append(os.path.join(root, name))

        found_entries = set()
        for entry in entries:
            relpath = os.path.relpath(entry, self.opts.work_dir)

            if os.path.isdir(entry):
                relpath += os.sep

            found_entries.add(relpath.replace(os.sep, posixpath.sep))

        required_set = set(required)
        missing = sorted(required_set.difference(found_entries))
        unknowns = sorted(found_entries.difference(required))

        if missing or unknowns:
            print('failed to verify extracted contents for: {}'.format(
                self.opts.cache_file))
            print('')

        if missing:
            print('missing entries]')
            for entry in missing:
                print(entry)
            print('')

        if unknowns:
            print('unknown entries]')
            for entry in unknowns:
                print(entry)
            print('')

        self.assertTrue(not missing and not unknowns)



# -*- coding: utf-8 -*-
# Copyright 2018-2022 releng-tool

from releng_tool.tool.tar import TAR
from releng_tool.util.io import ensure_dir_exists
from releng_tool.util.io import execute
from releng_tool.util.io import interpret_stem_extension
from releng_tool.util.log import debug
from releng_tool.util.log import err
from releng_tool.util.log import warn
from zipfile import ZipFile
import os
import shutil
import tarfile

#: list of (most-likely) supported tar extraction extensions
TAR_SUPPORTED = (
    'tar',
    'tb2',
    'tbz2',
    'tgz',
)

def extract(opts):
    """
    support extraction of an archive into a build directory

    With provided extraction options (``RelengExtractOptions``), the extraction
    stage will be processed. The archive's extension will be used in attempt to
    finding a matching tool/implementation which can be used to extract the
    contents of the file. In the event that the method of extraction cannot be
    determined, it will be assumed that the file is in fact not extractable.
    Files which are not extracted are just copied into the build directly (e.g.
    single resource files).

    Args:
        opts: the extraction options

    Returns:
        ``True`` if the extraction stage is completed; ``False`` otherwise
    """

    assert opts
    cache_file = opts.cache_file
    strip_count = opts.strip_count
    work_dir = opts.work_dir

    cache_basename = os.path.basename(cache_file)
    __, cache_ext = interpret_stem_extension(cache_basename)

    is_extractable = False
    if cache_ext:
        cache_ext = cache_ext.lower()

        # if the user defines a tool override for this extension type, use
        # whatever the user wants to use (passing the file and directory to
        # extract to)
        extract_override = getattr(opts, '_extract_override', None)
        if extract_override and cache_ext in extract_override:
            is_extractable = True

            tool_cmd = extract_override[cache_ext].format(
                file=cache_file, dir=work_dir)

            if not execute(tool_cmd.split(), cwd=work_dir, critical=False):
                err('unable to extract with tool override\n'
                    ' (command: {})', tool_cmd)
                return None

        # attempt to extract the (compressed) tar archive with the host's
        # tar tool; if it does not exist, we'll fallback to using python's
        # internal implementation (tarfile)
        elif cache_ext.startswith(TAR_SUPPORTED):
            is_extractable = True

            # before attempting to use an external tar command, only allow
            # using it if the `force-local` option is available whenever a
            # colon character is provided, to prevent tar from assuming the
            # path is a remote target
            needs_force_local = False
            if ':' in cache_file:
                needs_force_local = True

            has_extracted = False
            if TAR.exists() and (TAR.force_local or not needs_force_local):
                tar_args = [
                    '--extract',
                    '--file=' + cache_file,
                    '--strip-components={}'.format(strip_count),
                    '--verbose',
                ]

                if needs_force_local:
                    tar_args.append('--force-local')

                if TAR.execute(tar_args, cwd=work_dir):
                    has_extracted = True
                else:
                    warn('unable to extract archive with host tar; '
                        'will use fallback')

            if not has_extracted:
                try:
                    def tar_extract(members, strip_count):
                        for member in members:
                            # strip members from package defined count
                            if strip_count > 0:
                                np = os.path.normpath(member.name)
                                parts = np.split(os.path.sep, strip_count)
                                if len(parts) <= strip_count:
                                    continue
                                member.name = parts[-1]

                            # notify the user of the target member to extract
                            print(member.name)
                            yield member

                    with tarfile.open(cache_file, 'r') as tar:
                        tar.extractall(path=work_dir,
                            members=tar_extract(tar, strip_count))
                except Exception as e:
                    err('unable to extract tar file\n'
                        '    {}\n'
                        ' (file: {})\n'
                        ' (target: {})', e, cache_file, work_dir)
                    return False

        # extract a zip-extension cache file using python's internal
        # implementation (zipfile)
        elif cache_ext == 'zip':
            is_extractable = True

            try:
                with ZipFile(cache_file, 'r') as zip_:
                    for member in zip_.namelist():
                        # strip members from package defined count
                        member_s = member
                        if strip_count > 0:
                            np = os.path.normpath(member_s)
                            parts = np.split(os.path.sep, strip_count)
                            if len(parts) <= strip_count:
                                continue
                            member_s = parts[-1]
                        dest = os.path.join(work_dir, member_s)

                        # notify the user of the target member to extract
                        print(member)

                        # if this is a directory entry, ensure the directory
                        # exists for the destination
                        if not os.path.basename(member):
                            ensure_dir_exists(dest)
                        else:
                            # always ensure the container directory for a file
                            # exists before attempting to extract a member into
                            # it, as not all processed zip files may process
                            # a directory entry (to be created) ahead of time
                            ensure_dir_exists(os.path.dirname(dest))

                            with zip_.open(member) as s, open(dest, 'wb') as f:
                                shutil.copyfileobj(s, f)

            except Exception as e:
                err('unable to extract zip file\n'
                    '    {}\n'
                    ' (file: {})\n'
                    ' (target: {})', e, cache_file, work_dir)
                return False

    if not is_extractable:
        debug('file not considered extractable: ' + cache_file)
        try:
            shutil.copy2(cache_file, work_dir)
        except IOError as e:
            err('unable to copy over cache file\n'
                '    {}\n'
                ' (file: {})\n'
                ' (target: {})', e, cache_file, work_dir)
            return False

    return True



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

from releng_tool.opts import RelengEngineOptions
from releng_tool.packages.manager import RelengPackageManager
from releng_tool.registry import RelengRegistry
from tests.support import fetch_unittest_assets_dir
import os
import unittest

# base directory in the unit test folder which holds configurations
CONFIGS_DIR = 'configs'

class TestPkgConfigsBase(unittest.TestCase):
    """
    package configuration unit test base

    Provides a base class for unit testing package configurations. Using this
    base allows prepares a package manager and provides a quick configuration
    loader for configurations stored in the unit test's example configuration
    directory.

    Attributes:
        configs_dir: the configuration directory
        manager: the package manager
    """

    @classmethod
    def setUpClass(cls):
        opts = RelengEngineOptions()
        registry = RelengRegistry()
        cls.manager = RelengPackageManager(opts, registry)

        assets_dir = fetch_unittest_assets_dir()
        cls.configs_dir = os.path.join(assets_dir, CONFIGS_DIR)

    def LOAD(self, script_name, manager=None):
        """
        attempts to load a package definition

        With a provided `script_name`, the script will be loaded and will
        populate/return package related information.

        Args:
            script_name: the script to load from the configuration directory
            manager (optional): override the package manager to use when loading

        Returns:
            returns a tuple of three (3) containing the package instance, the
            extracted environment/globals from the package script and a list of
            known package dependencies

        Raises:
            RelengToolInvalidPackageConfiguration: when an error has been
                                                    detected loading the package
        """
        if not manager:
            manager = self.manager
        script = os.path.join(self.configs_dir, script_name)
        return manager.load_package('test', script)



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

from releng_tool.packages.exceptions import RelengToolInvalidPackageKeyValue
from tests.support.pkg_config_test import TestPkgConfigsBase

class TestPkgConfigsStageOpts(TestPkgConfigsBase):
    def test_pkgconfig_build_defs_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('build-defs-invalid-base-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('build-defs-invalid-key-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('build-defs-invalid-strs')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('build-defs-invalid-value-type')

    def test_pkgconfig_build_defs_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.build_defs)

        pkg, _, _ = self.LOAD('conf-defs-valid')
        self.assertIsNone(pkg.build_defs)

        pkg, _, _ = self.LOAD('install-defs-valid')
        self.assertIsNone(pkg.build_defs)

    def test_pkgconfig_build_defs_valid(self):
        pkg, _, _ = self.LOAD('build-defs-valid')
        self.assertDictEqual(pkg.build_defs, {
            'key1': 'val1',
            'key2': None,
            'key3': 'val3',
        })

    def test_pkgconfig_build_env_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('build-env-invalid-base-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('build-env-invalid-key-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('build-env-invalid-strs')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('build-env-invalid-value-type')

    def test_pkgconfig_build_env_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.build_env)

        pkg, _, _ = self.LOAD('conf-env-valid')
        self.assertIsNone(pkg.build_env)

        pkg, _, _ = self.LOAD('install-env-valid')
        self.assertIsNone(pkg.build_env)

    def test_pkgconfig_build_env_valid(self):
        pkg, _, _ = self.LOAD('build-env-valid')
        self.assertDictEqual(pkg.build_env, {
            'key1': 'val1',
            'key2': None,
            'key3': 'val3',
        })

    def test_pkgconfig_build_opts_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('build-opts-invalid-base-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('build-opts-invalid-key-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('build-opts-invalid-value-type')

    def test_pkgconfig_build_opts_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.build_opts)

        pkg, _, _ = self.LOAD('conf-opts-valid-dict')
        self.assertIsNone(pkg.build_opts)

        pkg, _, _ = self.LOAD('conf-opts-valid-str')
        self.assertIsNone(pkg.build_opts)

        pkg, _, _ = self.LOAD('conf-opts-valid-strs')
        self.assertIsNone(pkg.build_opts)

        pkg, _, _ = self.LOAD('install-opts-valid-dict')
        self.assertIsNone(pkg.build_opts)

        pkg, _, _ = self.LOAD('install-opts-valid-str')
        self.assertIsNone(pkg.build_opts)

        pkg, _, _ = self.LOAD('install-opts-valid-strs')
        self.assertIsNone(pkg.build_opts)

    def test_pkgconfig_build_opts_valid(self):
        pkg, _, _ = self.LOAD('build-opts-valid-dict')
        self.assertDictEqual(pkg.build_opts, {
            'key1': 'val1',
            'key2': None,
            'key3': 'val3',
        })

        pkg, _, _ = self.LOAD('build-opts-valid-str')
        self.assertDictEqual(pkg.build_opts, {
            'option': '',
        })

        pkg, _, _ = self.LOAD('build-opts-valid-strs')
        self.assertDictEqual(pkg.build_opts, {
            'option1': '',
            'option2': '',
            'option3': '',
        })

    def test_pkgconfig_conf_defs_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('conf-defs-invalid-base-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('conf-defs-invalid-key-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('conf-defs-invalid-strs')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('conf-defs-invalid-value-type')

    def test_pkgconfig_conf_defs_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.conf_defs)

        pkg, _, _ = self.LOAD('build-defs-valid')
        self.assertIsNone(pkg.conf_defs)

        pkg, _, _ = self.LOAD('install-defs-valid')
        self.assertIsNone(pkg.conf_defs)

    def test_pkgconfig_conf_defs_valid(self):
        pkg, _, _ = self.LOAD('conf-defs-valid')
        self.assertDictEqual(pkg.conf_defs, {
            'key1': 'val1',
            'key2': None,
            'key3': 'val3',
        })

    def test_pkgconfig_conf_env_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('conf-env-invalid-base-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('conf-env-invalid-key-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('conf-env-invalid-strs')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('conf-env-invalid-value-type')

    def test_pkgconfig_conf_env_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.conf_env)

        pkg, _, _ = self.LOAD('build-env-valid')
        self.assertIsNone(pkg.conf_env)

        pkg, _, _ = self.LOAD('install-env-valid')
        self.assertIsNone(pkg.conf_env)

    def test_pkgconfig_conf_env_valid(self):
        pkg, _, _ = self.LOAD('conf-env-valid')
        self.assertDictEqual(pkg.conf_env, {
            'key1': 'val1',
            'key2': None,
            'key3': 'val3',
        })

    def test_pkgconfig_conf_opts_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('conf-opts-invalid-base-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('conf-opts-invalid-key-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('conf-opts-invalid-value-type')

    def test_pkgconfig_conf_opts_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.conf_opts)

        pkg, _, _ = self.LOAD('build-opts-valid-dict')
        self.assertIsNone(pkg.conf_opts)

        pkg, _, _ = self.LOAD('build-opts-valid-str')
        self.assertIsNone(pkg.conf_opts)

        pkg, _, _ = self.LOAD('build-opts-valid-strs')
        self.assertIsNone(pkg.conf_opts)

        pkg, _, _ = self.LOAD('install-opts-valid-dict')
        self.assertIsNone(pkg.conf_opts)

        pkg, _, _ = self.LOAD('install-opts-valid-str')
        self.assertIsNone(pkg.conf_opts)

        pkg, _, _ = self.LOAD('install-opts-valid-strs')
        self.assertIsNone(pkg.conf_opts)

    def test_pkgconfig_conf_opts_valid(self):
        pkg, _, _ = self.LOAD('conf-opts-valid-dict')
        self.assertDictEqual(pkg.conf_opts, {
            'key1': 'val1',
            'key2': None,
            'key3': 'val3',
        })

        pkg, _, _ = self.LOAD('conf-opts-valid-str')
        self.assertDictEqual(pkg.conf_opts, {
            'option': '',
        })

        pkg, _, _ = self.LOAD('conf-opts-valid-strs')
        self.assertDictEqual(pkg.conf_opts, {
            'option1': '',
            'option2': '',
            'option3': '',
        })

    def test_pkgconfig_install_defs_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('install-defs-invalid-base-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('install-defs-invalid-key-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('install-defs-invalid-strs')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('install-defs-invalid-value-type')

    def test_pkgconfig_install_defs_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.install_defs)

        pkg, _, _ = self.LOAD('build-defs-valid')
        self.assertIsNone(pkg.install_defs)

        pkg, _, _ = self.LOAD('conf-defs-valid')
        self.assertIsNone(pkg.install_defs)

    def test_pkgconfig_install_defs_valid(self):
        pkg, _, _ = self.LOAD('install-defs-valid')
        self.assertDictEqual(pkg.install_defs, {
            'key1': 'val1',
            'key2': None,
            'key3': 'val3',
        })

    def test_pkgconfig_install_env_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('install-env-invalid-base-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('install-env-invalid-key-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('install-env-invalid-strs')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('install-env-invalid-value-type')

    def test_pkgconfig_install_env_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.install_env)

        pkg, _, _ = self.LOAD('build-env-valid')
        self.assertIsNone(pkg.install_env)

        pkg, _, _ = self.LOAD('conf-env-valid')
        self.assertIsNone(pkg.install_env)

    def test_pkgconfig_install_env_valid(self):
        pkg, _, _ = self.LOAD('install-env-valid')
        self.assertDictEqual(pkg.install_env, {
            'key1': 'val1',
            'key2': None,
            'key3': 'val3',
        })

    def test_pkgconfig_install_opts_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('install-opts-invalid-base-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('install-opts-invalid-key-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('install-opts-invalid-value-type')

    def test_pkgconfig_install_opts_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.install_opts)

        pkg, _, _ = self.LOAD('build-opts-valid-dict')
        self.assertIsNone(pkg.install_opts)

        pkg, _, _ = self.LOAD('build-opts-valid-str')
        self.assertIsNone(pkg.install_opts)

        pkg, _, _ = self.LOAD('build-opts-valid-strs')
        self.assertIsNone(pkg.install_opts)

        pkg, _, _ = self.LOAD('conf-opts-valid-dict')
        self.assertIsNone(pkg.install_opts)

        pkg, _, _ = self.LOAD('conf-opts-valid-str')
        self.assertIsNone(pkg.install_opts)

        pkg, _, _ = self.LOAD('conf-opts-valid-strs')
        self.assertIsNone(pkg.install_opts)

    def test_pkgconfig_install_opts_valid(self):
        pkg, _, _ = self.LOAD('install-opts-valid-dict')
        self.assertDictEqual(pkg.install_opts, {
            'key1': 'val1',
            'key2': None,
            'key3': 'val3',
        })

        pkg, _, _ = self.LOAD('install-opts-valid-str')
        self.assertDictEqual(pkg.install_opts, {
            'option': '',
        })

        pkg, _, _ = self.LOAD('install-opts-valid-strs')
        self.assertDictEqual(pkg.install_opts, {
            'option1': '',
            'option2': '',
            'option3': '',
        })



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

from releng_tool.defs import PackageInstallType
from releng_tool.opts import RelengEngineOptions
from releng_tool.packages.exceptions import RelengToolConflictingConfiguration
from releng_tool.packages.exceptions import RelengToolInvalidPackageKeyValue
from releng_tool.packages.exceptions import RelengToolUnknownInstallType
from releng_tool.packages.manager import RelengPackageManager
from releng_tool.registry import RelengRegistry
from tests.support.pkg_config_test import TestPkgConfigsBase

class TestPkgConfigs(TestPkgConfigsBase):
    def test_pkgconfig_deps_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('deps-invalid-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('deps-invalid-value')

    def test_pkgconfig_deps_missing(self):
        _, _, deps = self.LOAD('missing')
        self.assertListEqual(deps, [])

    def test_pkgconfig_deps_valid(self):
        _, _, deps = self.LOAD('deps-valid-empty')
        self.assertListEqual(deps, [])

        _, _, deps = self.LOAD('deps-valid-multiple')
        self.assertListEqual(deps, [
            'dep1',
            'dep2',
            'dep3',
        ])

        _, _, deps = self.LOAD('deps-valid-single')
        self.assertListEqual(deps, [
            'dep',
        ])

    def test_pkgconfig_fetch_opts_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('fetch-opts-invalid-base-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('fetch-opts-invalid-key-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('fetch-opts-invalid-value-type')

    def test_pkgconfig_fetch_opts_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.fetch_opts)

    def test_pkgconfig_fetch_opts_valid(self):
        pkg, _, _ = self.LOAD('fetch-opts-valid-dict')
        self.assertDictEqual(pkg.fetch_opts, {
            'key1': None,
            'key4': 'value',
            'key8': None,
        })

        pkg, _, _ = self.LOAD('fetch-opts-valid-str')
        self.assertDictEqual(pkg.fetch_opts, {
            '--my-option': '',
        })

        pkg, _, _ = self.LOAD('fetch-opts-valid-strs')
        self.assertDictEqual(pkg.fetch_opts, {
            'option4': '',
            'option5': '',
            'option6': '',
        })

    def test_pkgconfig_fixed_jobs_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('fixed-jobs-invalid-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('fixed-jobs-invalid-value')

    def test_pkgconfig_fixed_jobs_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.fixed_jobs)

    def test_pkgconfig_fixed_jobs_valid(self):
        pkg, _, _ = self.LOAD('fixed-jobs-valid')
        self.assertEqual(pkg.fixed_jobs, 24)

    def test_pkgconfig_install_type_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('install-type-invalid-type')

        with self.assertRaises(RelengToolUnknownInstallType):
            self.LOAD('install-type-invalid-value')

    def test_pkgconfig_install_type_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.install_type)

    def test_pkgconfig_install_type_valid(self):
        pkg, _, _ = self.LOAD('install-type-valid-host')
        self.assertEqual(pkg.install_type, PackageInstallType.HOST)

        pkg, _, _ = self.LOAD('install-type-valid-images')
        self.assertEqual(pkg.install_type, PackageInstallType.IMAGES)

        pkg, _, _ = self.LOAD('install-type-valid-staging')
        self.assertEqual(pkg.install_type, PackageInstallType.STAGING)

        pkg, _, _ = self.LOAD('install-type-valid-staging-target')
        self.assertEqual(pkg.install_type,
            PackageInstallType.STAGING_AND_TARGET)

        pkg, _, _ = self.LOAD('install-type-valid-target')
        self.assertEqual(pkg.install_type, PackageInstallType.TARGET)

    def test_pkgconfig_is_internal_disabled(self):
        pkg, _, _ = self.LOAD('internal-flag-disabled')
        self.assertFalse(pkg.is_internal)

        pkg, _, _ = self.LOAD('external-flag-disabled')
        self.assertTrue(pkg.is_internal)

        pkg, _, _ = self.LOAD('internal-external-flag-disabled')
        self.assertFalse(pkg.is_internal)

    def test_pkgconfig_is_internal_enabled(self):
        pkg, _, _ = self.LOAD('internal-flag-enabled')
        self.assertTrue(pkg.is_internal)

        pkg, _, _ = self.LOAD('external-flag-enabled')
        self.assertFalse(pkg.is_internal)

        pkg, _, _ = self.LOAD('internal-external-flag-enabled')
        self.assertTrue(pkg.is_internal)

    def test_pkgconfig_is_internal_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('internal-flag-invalid')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('external-flag-invalid')

        with self.assertRaises(RelengToolConflictingConfiguration):
            self.LOAD('internal-external-flag-conflict1')

        with self.assertRaises(RelengToolConflictingConfiguration):
            self.LOAD('internal-external-flag-conflict2')

    def test_pkgconfig_is_internal_missing_default(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertFalse(pkg.is_internal)

    def test_pkgconfig_is_internal_missing_implicit_internal(self):
        # force engine options to default packages to internal
        opts = RelengEngineOptions()
        opts.default_internal_pkgs = True

        registry = RelengRegistry()
        manager = RelengPackageManager(opts, registry)

        pkg, _, _ = self.LOAD('missing', manager=manager)
        self.assertTrue(pkg.is_internal)

    def test_pkgconfig_license_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('license-invalid-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('license-invalid-value')

    def test_pkgconfig_license_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.license)

    def test_pkgconfig_license_valid(self):
        pkg, _, _ = self.LOAD('license-valid-empty')
        self.assertListEqual(pkg.license, [])

        pkg, _, _ = self.LOAD('license-valid-multiple')
        self.assertListEqual(pkg.license, [
            'license1',
            'license2',
            'license3',
        ])

        pkg, _, _ = self.LOAD('license-valid-single')
        self.assertListEqual(pkg.license, [
            'license',
        ])

    def test_pkgconfig_license_files_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('license-files-invalid-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('license-files-invalid-value')

    def test_pkgconfig_license_files_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.license_files)

    def test_pkgconfig_license_files_valid(self):
        pkg, _, _ = self.LOAD('license-files-valid-empty')
        self.assertListEqual(pkg.license_files, [])

        pkg, _, _ = self.LOAD('license-files-valid-multiple')
        self.assertListEqual(pkg.license_files, [
            'license-file1',
            'license-file2',
            'license-file3',
        ])

        pkg, _, _ = self.LOAD('license-files-valid-single')
        self.assertListEqual(pkg.license_files, [
            'license-file',
        ])

    def test_pkgconfig_no_extraction_disabled(self):
        pkg, _, _ = self.LOAD('no-extraction-disabled')
        self.assertFalse(pkg.no_extraction)

    def test_pkgconfig_no_extraction_enabled(self):
        pkg, _, _ = self.LOAD('no-extraction-enabled')
        self.assertTrue(pkg.no_extraction)

    def test_pkgconfig_no_extraction_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('no-extraction-invalid')

    def test_pkgconfig_no_extraction_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.no_extraction)

    def test_pkgconfig_prefix_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('prefix-invalid-type')

    def test_pkgconfig_prefix_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.prefix)

    def test_pkgconfig_prefix_valid(self):
        pkg, _, _ = self.LOAD('prefix-valid')
        self.assertEqual(pkg.prefix, 'myprefix')

    def test_pkgconfig_python_interpreter_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('python-interpreter-invalid-type')

    def test_pkgconfig_python_interpreter_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.python_interpreter)

    def test_pkgconfig_python_interpreter_valid(self):
        pkg, _, _ = self.LOAD('python-interpreter-valid')
        self.assertEqual(pkg.python_interpreter, 'mypython')

    def test_pkgconfig_revision_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('revision-invalid-type')

    def test_pkgconfig_revision_valid_explicit(self):
        pkg, _, _ = self.LOAD('revision-valid-explicit')
        self.assertEqual(pkg.revision, 'myrevision')

    def test_pkgconfig_revision_valid_implicit(self):
        pkg, _, _ = self.LOAD('revision-valid-implicit')
        self.assertEqual(pkg.revision, 'myversion')

    def test_pkgconfig_site_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('site-invalid-type')

    def test_pkgconfig_site_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.site)

    def test_pkgconfig_site_valid(self):
        pkg, _, _ = self.LOAD('site-valid')
        self.assertEqual(pkg.site, 'https://example.com/file.tgz')

    def test_pkgconfig_skip_remote_config_disabled(self):
        pkg, _, _ = self.LOAD('skip-remote-config-disabled')
        self.assertFalse(pkg.skip_remote_config)

    def test_pkgconfig_skip_remote_config_enabled(self):
        pkg, _, _ = self.LOAD('skip-remote-config-enabled')
        self.assertTrue(pkg.skip_remote_config)

    def test_pkgconfig_skip_remote_config_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('skip-remote-config-invalid')

    def test_pkgconfig_skip_remote_config_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.skip_remote_config)

    def test_pkgconfig_skip_remote_scripts_disabled(self):
        pkg, _, _ = self.LOAD('skip-remote-scripts-disabled')
        self.assertFalse(pkg.skip_remote_scripts)

    def test_pkgconfig_skip_remote_scripts_enabled(self):
        pkg, _, _ = self.LOAD('skip-remote-scripts-enabled')
        self.assertTrue(pkg.skip_remote_scripts)

    def test_pkgconfig_skip_remote_scripts_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('skip-remote-scripts-invalid')

    def test_pkgconfig_skip_remote_scripts_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertIsNone(pkg.skip_remote_scripts)

    def test_pkgconfig_strip_count_invalid(self):
        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('strip-count-invalid-type')

        with self.assertRaises(RelengToolInvalidPackageKeyValue):
            self.LOAD('strip-count-invalid-value')

    def test_pkgconfig_strip_count_missing(self):
        pkg, _, _ = self.LOAD('missing')
        self.assertEqual(pkg.strip_count, 1) # default

    def test_pkgconfig_strip_count_valid(self):
        pkg, _, _ = self.LOAD('strip-count-valid-zero')
        self.assertEqual(pkg.strip_count, 0)

        pkg, _, _ = self.LOAD('strip-count-valid-nonzero')
        self.assertEqual(pkg.strip_count, 2)



# -*- coding: utf-8 -*-
# Copyright 2020 releng-tool
# flake8: noqa

from releng_tool import *

if '_RELENG_GLOBAL_NAMESPACE_DEPRECATED' not in globals():
    global _RELENG_GLOBAL_NAMESPACE_DEPRECATED
    _RELENG_GLOBAL_NAMESPACE_DEPRECATED = True
    warn('(deprecated) namespace releng has been deprecated over releng_tool')



# -*- coding: utf-8 -*-
# Copyright 2021-2022 releng-tool

from contextlib import contextmanager
from releng_tool.defs import GlobalAction
from releng_tool.defs import PkgAction
from releng_tool.defs import VcsType
from releng_tool.engine.bootstrap import stage as bootstrap_stage
from releng_tool.engine.build import stage as build_stage
from releng_tool.engine.configure import stage as configure_stage
from releng_tool.engine.extract import stage as extract_stage
from releng_tool.engine.install import stage as install_stage
from releng_tool.engine.patch import stage as patch_stage
from releng_tool.engine.post import stage as post_stage
from releng_tool.exceptions import RelengToolMissingExecCommand
from releng_tool.packages.exceptions import RelengToolBootstrapStageFailure
from releng_tool.packages.exceptions import RelengToolBuildStageFailure
from releng_tool.packages.exceptions import RelengToolConfigurationStageFailure
from releng_tool.packages.exceptions import RelengToolExecStageFailure
from releng_tool.packages.exceptions import RelengToolExtractionStageFailure
from releng_tool.packages.exceptions import RelengToolInstallStageFailure
from releng_tool.packages.exceptions import RelengToolLicenseStageFailure
from releng_tool.packages.exceptions import RelengToolPatchStageFailure
from releng_tool.packages.exceptions import RelengToolPostStageFailure
from releng_tool.util.file_flags import FileFlag
from releng_tool.util.file_flags import check_file_flag
from releng_tool.util.file_flags import process_file_flag
from releng_tool.util.io import ensure_dir_exists
from releng_tool.util.io import path_copy
from releng_tool.util.log import debug
from releng_tool.util.log import err
from releng_tool.util.log import note
from releng_tool.util.log import warn
import os
import subprocess
import sys

class RelengPackagePipeline:
    """
    a releng package pipeline

    A pipeline will process a provide package through the various stages of its
    life (extracting, patching, configuring, etc.).

    Args:
        engine: the package engine
        opts: options used to configure the engine
        script_env: the script environment to load

    Attributes:
        license_files: dictionary of package-to-license entities found when
                        processing packages
        opts: options used to configure the engine
        script_env: the script environment to load
    """
    def __init__(self, engine, opts, script_env):
        self.engine = engine
        self.license_files = {}
        self.opts = opts
        self.script_env = script_env

    def process(self, pkg):
        """
        process a provided package

        This request will process a package through the various stages (if these
        stages are applicable to the current run state and have yet been
        executed from a previous run). A package-specific script environment
        will be prepared and a package will go through the process of:

        - Extraction
        - Patching
        - License management (if needed)
        - Bootstrapping
        - Configuration
        - Building
        - Installing
        - Post-processing

        Args:
            pkg: the package to process

        Returns:
            returns whether or not the pipeline should continue processing
        """
        gaction = self.opts.gbl_action
        paction = self.opts.pkg_action
        target = self.opts.target_action

        # skip if generating license information and no license
        # files exist for this package
        if gaction == GlobalAction.LICENSES and not pkg.license_files:
            return True

        # prepare environment
        pkg_env = self._stage_env(pkg)

        # extracting
        flag = pkg._ff_extract
        if check_file_flag(flag) == FileFlag.NO_EXIST:
            self.engine.stats.track_duration_start(pkg.name, 'extract')
            # none/local-vcs-type packages do not need to fetch
            if pkg.vcs_type in (VcsType.LOCAL, VcsType.NONE):
                pass
            elif not extract_stage(self.engine, pkg):
                raise RelengToolExtractionStageFailure
            # now that the extraction stage has (most likely)
            # created a build directory, ensure the output directory
            # exists as well (for file flags and other content)
            if not ensure_dir_exists(pkg.build_output_dir):
                raise RelengToolExtractionStageFailure
            self.engine.stats.track_duration_end(pkg.name, 'extract')
            if process_file_flag(True, flag) != FileFlag.CONFIGURED:
                return False
        if gaction == GlobalAction.EXTRACT:
            return True
        if paction == PkgAction.EXTRACT and pkg.name == target:
            return False

        # patching
        flag = pkg._ff_patch
        if check_file_flag(flag) == FileFlag.NO_EXIST:
            self.engine.stats.track_duration_start(pkg.name, 'patch')
            # local-vcs-type packages do not need to patch
            if pkg.vcs_type is VcsType.LOCAL:
                pass
            elif not patch_stage(self.engine, pkg, pkg_env):
                raise RelengToolPatchStageFailure
            self.engine.stats.track_duration_end(pkg.name, 'patch')
            if process_file_flag(True, flag) != FileFlag.CONFIGURED:
                return False
        if gaction == GlobalAction.PATCH:
            return True
        if paction == PkgAction.PATCH and pkg.name == target:
            return False

        # handle license generation request
        #
        # If the user has requested to generate license information,
        # pull license assets from the extract package content.
        # license(s)
        flag = pkg._ff_license
        if check_file_flag(flag) == FileFlag.NO_EXIST:
            if not self._stage_license(pkg):
                raise RelengToolLicenseStageFailure
            if process_file_flag(True, flag) != FileFlag.CONFIGURED:
                return False

        if pkg.license_files:
            version_desc = pkg.version
            if not version_desc and pkg.revision:
                version_desc = pkg.revision

            self.license_files[pkg.name] = {
                'files': [],
                'version': version_desc,
            }
            for file in pkg.license_files:
                file = os.path.join(pkg.build_dir, file)
                self.license_files[pkg.name]['files'].append(file)

        if gaction == GlobalAction.LICENSES:
            return True
        if paction == PkgAction.LICENSE and pkg.name == target:
            return False

        # load any late-stage configuration options from the remote
        # sources
        if (not pkg.skip_remote_config and
                'releng.disable_remote_configs' not in self.opts.quirks):
            self.engine.pkgman.load_remote_configuration(pkg)

        # finalize package environment
        with self._stage_env_finalize(pkg, pkg_env):

            # custom execution command
            if paction == PkgAction.EXEC and pkg.name == target:
                self._stage_exec(pkg)
                return False

            # bootstrapping
            flag = pkg._ff_bootstrap
            if check_file_flag(flag) == FileFlag.NO_EXIST:
                self.engine.stats.track_duration_start(pkg.name, 'boot')
                if not bootstrap_stage(self.engine, pkg, pkg_env):
                    raise RelengToolBootstrapStageFailure
                self.engine.stats.track_duration_end(pkg.name, 'boot')
                if process_file_flag(True, flag) != FileFlag.CONFIGURED:
                    return False

            # configuring
            flag = pkg._ff_configure
            if check_file_flag(flag) == FileFlag.NO_EXIST:
                self.engine.stats.track_duration_start(pkg.name, 'configure')
                if not configure_stage(self.engine, pkg, pkg_env):
                    raise RelengToolConfigurationStageFailure
                self.engine.stats.track_duration_end(pkg.name, 'configure')
                if process_file_flag(True, flag) != FileFlag.CONFIGURED:
                    return False
            if paction in (PkgAction.CONFIGURE, PkgAction.RECONFIGURE_ONLY):
                if pkg.name == target:
                    return False

            # building
            flag = pkg._ff_build
            if check_file_flag(flag) == FileFlag.NO_EXIST:
                self.engine.stats.track_duration_start(pkg.name, 'build')
                if not build_stage(self.engine, pkg, pkg_env):
                    raise RelengToolBuildStageFailure
                self.engine.stats.track_duration_end(pkg.name, 'build')
                if process_file_flag(True, flag) != FileFlag.CONFIGURED:
                    return False
            if paction in (PkgAction.BUILD, PkgAction.REBUILD_ONLY):
                if pkg.name == target:
                    return False

            # installing
            flag = pkg._ff_install
            if check_file_flag(flag) == FileFlag.NO_EXIST:
                self.engine.stats.track_duration_start(pkg.name, 'install')
                if not install_stage(self.engine, pkg, pkg_env):
                    raise RelengToolInstallStageFailure
                self.engine.stats.track_duration_end(pkg.name, 'install')
                if process_file_flag(True, flag) != FileFlag.CONFIGURED:
                    return False
            # (note: re-install requests will re-invoke package-specific
            # post-processing)

            # package-specific post-processing
            flag = pkg._ff_post
            if check_file_flag(flag) == FileFlag.NO_EXIST:
                self.engine.stats.track_duration_start(pkg.name, 'post')
                if not post_stage(self.engine, pkg, pkg_env):
                    raise RelengToolPostStageFailure
                self.engine.stats.track_duration_end(pkg.name, 'post')
                if process_file_flag(True, flag) != FileFlag.CONFIGURED:
                    return False
            if paction in (
                    PkgAction.INSTALL,
                    PkgAction.REBUILD_ONLY,
                    PkgAction.RECONFIGURE_ONLY,
                    PkgAction.REINSTALL):
                if pkg.name == target:
                    return False

        return True

    def _stage_env(self, pkg):
        """
        prepare environment variables for a specific package processing

        When a package is being processed (configuration, building, etc.), a
        unique set of environment variables may be provided specifically for the
        package. These are provided out of convenience as an alternative to
        needing to rely on the Python-provided stage options.

        Args:
            pkg: the package being processed

        Returns:
            the prepared package-enhanced environment variables
        """

        # copy environment since packages do not share values
        pkg_env = self.script_env.copy()

        if pkg.build_subdir:
            build_dir = pkg.build_subdir
        else:
            build_dir = pkg.build_dir

        # always register optional flags in script environment
        pkg_env['PKG_INTERNAL'] = None

        # package variables
        for env in (os.environ, pkg_env):
            env['PKG_BUILD_BASE_DIR'] = pkg.build_dir
            env['PKG_BUILD_DIR'] = build_dir
            env['PKG_BUILD_OUTPUT_DIR'] = pkg.build_output_dir
            env['PKG_CACHE_DIR'] = pkg.cache_dir
            env['PKG_CACHE_FILE'] = pkg.cache_file
            env['PKG_DEFDIR'] = pkg.def_dir
            env['PKG_NAME'] = pkg.name
            env['PKG_SITE'] = pkg.site if pkg.site else ''
            env['PKG_REVISION'] = pkg.revision
            env['PKG_VERSION'] = pkg.version

            if pkg.is_internal:
                env['PKG_INTERNAL'] = '1'

        return pkg_env

    @contextmanager
    def _stage_env_finalize(self, pkg, pkg_env):
        """
        finalize environment variables for a specific package processing

        While a `_stage_env` call will prepare the script environment for
        various stages of a package, some options of a package may be loaded in
        at a later stage. Options which can be late-loaded are populated into
        the existing environment in this call.

        Args:
            pkg: the package being processed
            pkg_env: the environment to populate
        """

        pkg_keys = [
            'HOST_INCLUDE_DIR',
            'HOST_LIB_DIR',
            'NJOBS',
            'NJOBSCONF',
            'PREFIX',
            'PREFIXED_HOST_DIR',
            'PREFIXED_STAGING_DIR',
            'PREFIXED_TARGET_DIR',
            'STAGING_INCLUDE_DIR',
            'STAGING_LIB_DIR',
            'TARGET_INCLUDE_DIR',
            'TARGET_LIB_DIR',
        ]

        saved_env = {}
        for key in pkg_keys:
            saved_env[key] = os.environ.get(key, None)

        # apply package specific overrides into the OS environment and the
        # package/script environment
        try:
            for env in (os.environ, pkg_env):
                if pkg.prefix is not None:
                    nprefix = os.path.normpath(pkg.prefix)
                    host_pdir = self.opts.host_dir + nprefix
                    staging_pdir = self.opts.staging_dir + nprefix
                    target_pdir = self.opts.target_dir + nprefix

                    host_include_dir = os.path.join(host_pdir, 'include')
                    host_lib_dir = os.path.join(host_pdir, 'lib')
                    staging_include_dir = os.path.join(staging_pdir, 'include')
                    staging_lib_dir = os.path.join(staging_pdir, 'lib')
                    target_include_dir = os.path.join(target_pdir, 'include')
                    target_lib_dir = os.path.join(target_pdir, 'lib')

                    # will override existing prefix related variables
                    env['HOST_INCLUDE_DIR'] = host_include_dir
                    env['HOST_LIB_DIR'] = host_lib_dir
                    env['PREFIX'] = pkg.prefix
                    env['PREFIXED_HOST_DIR'] = host_pdir
                    env['PREFIXED_STAGING_DIR'] = staging_pdir
                    env['PREFIXED_TARGET_DIR'] = target_pdir
                    env['STAGING_INCLUDE_DIR'] = staging_include_dir
                    env['STAGING_LIB_DIR'] = staging_lib_dir
                    env['TARGET_INCLUDE_DIR'] = target_include_dir
                    env['TARGET_LIB_DIR'] = target_lib_dir

                if pkg.fixed_jobs:
                    env['NJOBS'] = str(pkg.fixed_jobs)
                    env['NJOBSCONF'] = str(pkg.fixed_jobs)

            yield
        finally:
            # restore any overrides that may have been set
            for k, v in saved_env.items():
                if v is not None:
                    os.environ[k] = v
                else:
                    os.environ.pop(k, None)

    def _stage_exec(self, pkg):
        """
        execute a command for a specific package

        Provides a user the ability to invoke a command in a package's
        extracted directory. This is a helper if a user wishes to invoke/test
        commands for a package without needing to navigate to the package's
        build directory and invoking them their.

        Args:
            pkg: the package being processed

        Raises:
            RelengToolExecStageFailure: when the command returns non-zero value
            RelengToolMissingExecCommand: when no command is provided
        """

        exec_cmd = self.opts.target_action_exec
        if not exec_cmd:
            raise RelengToolMissingExecCommand(pkg.name)

        note('execution for {}...', pkg.name)
        debug('dir: {}', pkg.build_tree)
        debug('cmd: {}', exec_cmd)
        sys.stdout.flush()

        proc = subprocess.Popen(
            exec_cmd,
            cwd=pkg.build_tree,
            shell=True,
        )
        proc.communicate()
        sys.stdout.flush()

        if proc.returncode != 0:
            raise RelengToolExecStageFailure

    def _stage_license(self, pkg):
        """
        process license files for a specific package processing

        If a package contains one or more files containing licenses information,
        this information will be populated in the package's license folder.

        Args:
            pkg: the package being processed

        Returns:
            ``True`` if the license information was copied; ``False`` if these
            license information could not be copied
        """

        # skip if package has no license files
        if not pkg.license_files:
            if pkg.license and not pkg.is_internal and not pkg.no_extraction:
                warn('package defines no license files: ' + pkg.name)
            return True

        # ensure package-specific license directory exists
        pkg_license_dir = os.path.join(self.opts.license_dir, pkg.nv)
        if not ensure_dir_exists(pkg_license_dir):
            return False

        # copy over each license files
        for file in pkg.license_files:
            src = os.path.join(pkg.build_dir, file)
            dst = os.path.join(pkg_license_dir, file)

            if not path_copy(src, dst, critical=False):
                err('unable to copy license information: ' + pkg.name)
                return False

        return True



# -*- coding: utf-8 -*-
# Copyright 2018-2022 releng-tool

from releng_tool import __version__ as releng_version
from releng_tool.engine import RelengEngine
from releng_tool.exceptions import RelengToolException
from releng_tool.opts import RelengEngineOptions
from releng_tool.util.log import debug
from releng_tool.util.log import err
from releng_tool.util.log import releng_log_configuration
from releng_tool.util.log import verbose
from releng_tool.util.log import warn
from releng_tool.util.win32 import enable_ansi as enable_ansi_win32
import argparse
import os
import sys

def main():
    """
    mainline

    The mainline for the releng tool.

    Returns:
        the exit code
    """
    retval = 1

    try:
        parser = argparse.ArgumentParser(
            prog='releng-tool', add_help=False, usage=usage())

        parser.add_argument('--assets-dir')
        parser.add_argument('--cache-dir')
        parser.add_argument('--debug', action='store_true')
        parser.add_argument('--development', '-D', action='store_true')
        parser.add_argument('--dl-dir')
        parser.add_argument('--force', '-F', action='store_true')
        parser.add_argument('--help', '-h', action='store_true')
        parser.add_argument('--help-quirks', action='store_true')
        parser.add_argument('--jobs', '-j', default=0, type=type_nonnegativeint)
        parser.add_argument('--local-sources', action='store_true')
        parser.add_argument('--nocolorout', action='store_true')
        parser.add_argument('--out-dir')
        parser.add_argument('--root-dir')
        parser.add_argument('--quirk', action='append')
        parser.add_argument('--verbose', '-V', action='store_true')
        parser.add_argument('--version', action='version',
            version='%(prog)s ' + releng_version)

        known_args = sys.argv[1:]
        forward_args = []
        idx = known_args.index('--') if '--' in known_args else -1
        if idx != -1:
            forward_args = known_args[idx + 1:]
            known_args = known_args[:idx]

        args, unknown_args = parser.parse_known_args(known_args)
        if args.help:
            print(usage())
            sys.exit(0)
        if args.help_quirks:
            print(usage_quirks())
            sys.exit(0)

        # force verbose messages if debugging is enabled
        if args.debug:
            args.verbose = True

        releng_log_configuration(args.debug, args.nocolorout, args.verbose)

        # toggle on ansi colors by default for commands
        if not args.nocolorout:
            os.environ['CLICOLOR_FORCE'] = '1'

            # support character sequences (for color output on win32 cmd)
            if sys.platform == 'win32':
                enable_ansi_win32()

        verbose('releng-tool {}', releng_version)
        debug('({})', __file__)

        # extract additional argument information:
        #  - pull the action value
        #  - pull "exec" command (if applicable)
        #  - key-value entries to be injected into the running
        #     script/working environment
        new_args, unknown_args = process_args(unknown_args)
        args.action = new_args['action']
        args.action_exec = new_args['exec']
        args.injected_kv = new_args['entries']

        # register any injected entry into the working environment right away
        for k, v in args.injected_kv.items():
            os.environ[k] = v

        if unknown_args:
            warn('unknown arguments: {}', ' '.join(unknown_args))

        if forward_args:
            debug('forwarded arguments: {}', ' '.join(forward_args))

        # warn if the *nix-based system is running as root; ill-formed projects
        # may attempt to modify the local system's root
        if sys.platform != 'win32':
            if os.geteuid() == 0: # pylint: disable=E1101
                if 'RELENG_IGNORE_RUNNING_AS_ROOT' not in os.environ:
                    warn('running as root; this may be unsafe')

        # prepare engine options
        opts = RelengEngineOptions(args=args, forward_args=forward_args)

        # create and start the engine
        engine = RelengEngine(opts)
        try:
            if engine.run():
                retval = 0
        except RelengToolException as e:
            err(e)
    except KeyboardInterrupt:
        print('')

    return retval

def process_args(args):
    """
    process arguments for an action and key-value entries for environments

    The following will process a remaining argument set for an provided action
    and entries which could be injected into the script/working environment.
    The goal is to support Make-styled variable assignment options from a
    command line, providing users a consistent way to override/set options no
    matter the platform.

    Args:
        args: the arguments to check for entries

    Returns:
        parsed argument options and the remaining/unknown arguments
    """

    action = None
    entries = {}
    exec_ = None
    needs_exec = False
    unknown_args = list(args)

    for arg in args:
        # always ignore option entries
        if arg.startswith('-'):
            continue

        if needs_exec:
            exec_ = arg.strip()
            unknown_args.remove(arg)
            debug('detected package-exec call: {}', exec_)
            needs_exec = False

        is_entry = False
        if '=' in arg:
            key, value = arg.split('=', 1)
            key = key.strip()
            if key:
                is_entry = True
                entries[key] = value.strip()
                unknown_args.remove(arg)
                debug('detected entry: {}={}', key, entries[key])

        # if this argument is not an entry and we haven't yet registered an
        # action yet, consider this argument the action
        if not action and not is_entry:
            action = arg
            unknown_args.remove(arg)
            debug('detected action: {}', action)

            # if this is assumed to be an `exec`-based package action, consume
            # the next non-kv entry as the expected command to invoke
            if action.endswith('-exec'):
                debug('assuming action is an exec call')
                needs_exec = True

    return {
        'action': action,
        'entries': entries,
        'exec': exec_,
    }, unknown_args

def type_nonnegativeint(value):
    """
    argparse type check for a non-negative integer

    Provides a type check for an argparse-provided argument value to ensure the
    value is a non-negative integer value.

    Args:
        value: the value to check

    Returns:
        the non-negative integer value

    Raises:
        argparse.ArgumentTypeError: detected an invalid non-negative value
    """
    val = int(value)
    if val < 0:
        raise argparse.ArgumentTypeError('invalid non-negative value')
    return val

def usage():
    """
    display the usage for this tool

    Returns a command line usage string for all options available by the releng
    tool.

    Returns:
        the usage string
    """
    return """releng-tool <options> [action]

(actions)
 clean                     clean the output directory
 distclean                 pristine clean including cache/download content
 extract                   extract all packages
 fetch                     fetch all packages
 init                      initialize a root with an example structure
 licenses                  generate license information for a project
 mrproper                  pristine clean of the releng project
 patch                     ensure all packages have done a patch stage
 <pkg>-build               perform build stage for the package
 <pkg>-clean               clean build directory for package
 <pkg>-configure           perform configure stage for the package
 <pkg>-distclean           pristine clean for package
 <pkg>-exec <cmd>          invoke a command in the package's directory
 <pkg>-extract             perform extract stage for the package
 <pkg>-fetch               perform fetch stage for the package
 <pkg>-install             perform install stage for the package
 <pkg>-license             generate license information for the package
 <pkg>-patch               perform patch stage for the package
 <pkg>-rebuild             force a re-build of a specific package
 <pkg>-rebuild-only        force a re-build of a specific package and stop
 <pkg>-reconfigure         force a re-configure of a specific package
 <pkg>-reconfigure-only    force a re-configure of a specific package and stop
 <pkg>-reinstall           force a re-install of a specific package

(common options)
 --assets-dir <dir>        container directory for download and vcs-cache
                            directories (e.g. <ASSETS_DIR>/cache)
 --cache-dir <dir>         directory for vcs-cache (default: <ROOT>/cache)
 --dl-dir <dir>            directory for download archives (default: <ROOT>/dl)
 -j, --jobs <jobs>         numbers of jobs to handle (default: 0; automatic)
 --out-dir <dir>           directory for output (builds, images, etc.)
                            (default: <ROOT>/output)
 --root-dir <dir>          directory to process a releng project
                            (default: working directory)

(mode options)
 -D, --development         enable development mode
 --local-sources           use development sources from a local path

(other)
 --debug                   show debug-related messages
 -F, --force               trigger a forced request
 -h, --help                show this help
 --help-quirks             show available quirks
 --nocolorout              explicitly disable colorized output
 --quirk <value>           inject in quirk into this run
 -V, --verbose             show additional messages
 --version                 show the version
"""

def usage_quirks():
    """
    display the available quirks in this tool

    Returns a string of available quirks in this tool.

    Returns:
        the quirks string
    """
    return """releng-tool quirks

releng.cmake.disable_parallel_option   disable parallel cmake
releng.disable_prerequisites_check     disable prerequisites check
releng.disable_remote_configs          disable remote configurations
releng.disable_remote_scripts          disable remote scripts
releng.git.no_depth                    disable depth-limits for git calls
releng.git.no_quick_fetch              disable quick-fetching for git calls
"""

if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

from tests import prepare_testenv
import os
import unittest

class TestPkgPipelineRemoteConfigs(unittest.TestCase):
    def test_pkg_pipeline_remote_cfg_disabled_option(self):
        with prepare_testenv(template='remote-cfg-disabled') as engine:
            engine.run()
            self._assertFileFlagExists(engine, 'jobs-0')

    def test_pkg_pipeline_remote_cfg_disabled_quirk(self):
        config = {
            'quirk': [
                'releng.disable_remote_configs',
            ],
        }
        with prepare_testenv(config=config, template='remote-cfg') as engine:
            engine.run()
            self._assertFileFlagExists(engine, 'jobs-0')

    def test_pkg_pipeline_remote_cfg_enabled(self):
        with prepare_testenv(template='remote-cfg') as engine:
            engine.run()
            self._assertFileFlagExists(engine, 'jobs-1')

    def test_pkg_pipeline_remote_cfg_override(self):
        with prepare_testenv(template='remote-cfg-override') as engine:
            engine.run()
            self._assertFileFlagExists(engine, 'jobs-2')

    def _assertFileFlagExists(self, engine, name):
        file_flag = os.path.join(engine.opts.target_dir, name)
        self.assertTrue(os.path.exists(file_flag))
